# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""NVIDIA HPC SDK building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion
import logging
import re
import posixpath

import hpccm.config
import hpccm.templates.downloader
import hpccm.templates.envvars
import hpccm.templates.rm

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.packages import packages
from hpccm.common import cpu_arch, linux_distro
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.environment import environment
from hpccm.primitives.shell import shell
from hpccm.toolchain import toolchain

class nvhpc(bb_base, hpccm.templates.downloader, hpccm.templates.envvars,
            hpccm.templates.rm):
    """The `nvhpc` building block downloads and installs the [NVIDIA HPC
    SDK](https://developer.nvidia.com/hpc-sdk).  By default, the
    NVIDIA HPC SDK is installed from a package repository.
    Alternatively the tar package can be downloaded by specifying the
    `tarball` parameter, or a local tar package may used instead by
    specifying the `package` parameter.

    You must agree to the [NVIDIA HPC SDK End-User License Agreement](https://docs.nvidia.com/hpc-sdk/eula) to use this
    building block.

    As a side effect, a toolchain is created containing the NVIDIA
    compilers.  The tool can be passed to other operations that want
    to build using the NVIDIA compilers.

    # Parameters

    cuda: The default CUDA version to configure.  The default is an
    empty value, i.e., use the latest version supported by the NVIDIA
    HPC SDK.  This value is ignored if installing from the package
    repository.

    cuda_multi: Boolean flag to specify whether the NVIDIA HPC SDK
    support for multiple CUDA versions should be installed.  The
    default value is `True`.

    environment: Boolean flag to specify whether the environment
    (`CPATH`, `LD_LIBRARY_PATH`, `MANPATH`, and `PATH`) should be
    modified to include the NVIDIA HPC SDK. The default is True.

    eula: By setting this value to `True`, you agree to the [NVIDIA HPC SDK End-User License Agreement](https://docs.nvidia.com/hpc-sdk/eula).
    The default value is `False`.

    extended_environment: Boolean flag to specify whether an extended
    set of environment variables should be defined.  If True, the
    following environment variables `CC`, `CPP`, `CXX`, `F77`, `F90`,
    and `FC`.  If False, then only `CPATH`, `LD_LIBRARY_PATH`,
    `MANPATH`, and `PATH` will be extended to include the NVIDIA HPC
    SDK.  The default value is `False`.

    mpi: Boolean flag to specify whether MPI should be included in the
    environment.  The default value is `True`.

    ospackages: List of OS packages to install prior to installing the
    NVIDIA HPC SDK.  The default value is `ca-certificates`, `gnupg`,
    and `wget` for Ubuntu, and `ca-certificates` for RHEL-based Linux
    distributions.  If not installing from the package repository,
    then for Ubuntu, the default values are `bc`, `debianutils`,
    `gcc`, `g++`, `gfortran`, `libatomic1`, `libnuma1`,
    `openssh-client`, and `wget`, and for RHEL-based Linux
    distributions, the default values are `bc`, `gcc`, `gcc-c++`,
    `gcc-gfortran`, `libatomic`, `numactl-libs`, `openssh-clients`,
    `wget`, and `which`.

    package: Path to the NVIDIA HPC SDK tar package file relative to
    the local build context.  The default value is empty.

    prefix: The top level install prefix.  The default value is
    `/opt/nvidia/hpc_sdk`.  This value is ignored when installing from
    the package repository.

    redist: The list of redistributable files to copy into the runtime
    stage.  The paths are relative to the `REDIST` directory and
    wildcards are supported.  The default is an empty list.

    tarball: Boolean flag to specify whether the NVIDIA HPC SDK should
    be installed by downloading the tar package file.  If False,
    install from the package repository.  The default is False.

    toolchain: The toolchain object to be used to configure the HPC
    SDK with a specific GNU toolchain.  The default is empty, i.e., use
    the default GNU toolchain.

    url: The location of the package that should be installed.  The default value is `https://developer.download.nvidia.com/hpc-sdk/nvhpc_X_Y_Z_cuda_multi.tar.gz`, where `X, `Y`, and `Z` are the year, version, and architecture whose values are automatically determined.

    version: The version of the HPC SDK to use.  Note when `package`
    is set the version is determined automatically from the package
    file name.  The default value is `22.9`.

    # Examples

    ```python
    nvhpc(eula=True)
    ```

    ```python
    nvhpc(eula=True, tarball=True)
    ```

    ```python
    nvhpc(eula=True,
          url='https://developer.download.nvidia.com/hpc-sdk/nvhpc_2020_207_Linux_x86_64_cuda_11.0.tar.gz')
    ```

    ```python
    nvhpc(eula=True,
          package='nvhpc_2020_207_Linux_x86_64_cuda_multi.tar.gz',
          redist=['compilers/lib/*'])
    ```

    ```python
    n = nvhpc(eula=True, ...)
    openmpi(..., toolchain=n.toolchain, ...)
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(nvhpc, self).__init__(**kwargs)

        self.__arch_directory = None # Filled in by __cpu_arch()
        self.__arch_label = '' # Filled in by __cpu_arch()
        self.__cuda_multi = kwargs.get('cuda_multi', True)
        self.__cuda_version = kwargs.get('cuda', None)
        self.__commands = [] # Filled in by __setup_tarball()

        # By setting this value to True, you agree to the NVIDIA HPC
        # SDK End-User License Agreement
        # (https://docs.nvidia.com/hpc-sdk/eula)
        self.__eula = kwargs.get('eula', False)

        self.__extended_environment = kwargs.get('extended_environment', False)
        self.__hpcx = kwargs.get('_hpcx', False)
        self.__mpi = kwargs.get('mpi', True)
        self.__nvhpc_package = '' # Filled in by __distro()
        self.__ospackages = kwargs.get('ospackages', [])
        self.__runtime_ospackages = [] # Filled in by __distro()
        self.__prefix = kwargs.get('prefix', '/opt/nvidia/hpc_sdk')
        self.__redist = kwargs.get('redist', [])
        self.__stdpar_cudacc = kwargs.get('stdpar_cudacc', None)
        self.__tarball = kwargs.get('tarball', False)
        self.__toolchain = kwargs.get('toolchain', None)
        self.__url = kwargs.get('url', None)
        self.__version = kwargs.get('version', '22.9')
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory
        self.__year = '' # Filled in by __get_version()

        self.toolchain = toolchain(CC='nvc', CXX='nvc++', F77='nvfortran',
                                   F90='nvfortran', FC='nvfortran')

        if StrictVersion(self.__version) >= StrictVersion('22.5'):
            self.__cuda_version_default = '11.7'
        elif StrictVersion(self.__version) >= StrictVersion('22.2'):
            self.__cuda_version_default = '11.6'
        elif StrictVersion(self.__version) >= StrictVersion('21.11'):
            self.__cuda_version_default = '11.5'
        elif StrictVersion(self.__version) >= StrictVersion('21.7'):
            self.__cuda_version_default = '11.4'
        elif StrictVersion(self.__version) >= StrictVersion('21.5'):
            self.__cuda_version_default = '11.3'
        elif StrictVersion(self.__version) >= StrictVersion('21.2'):
            self.__cuda_version_default = '11.2'
        elif StrictVersion(self.__version) >= StrictVersion('20.11'):
            self.__cuda_version_default = '11.1'
        else:
            self.__cuda_version_default = '11.0'

        # Set the CPU architecture specific parameters
        self.__cpu_arch()

        # Set the Linux distribution specific parameters
        self.__distro()

        # Figure out the version information
        self.__get_version()

        # Set paths used extensively
        self.__basepath = posixpath.join(self.__prefix, self.__arch_directory,
                                         self.__version)

        # Construct the series of steps to execute
        if self.package or self.__tarball or self.__url:
            self.__setup_tarball()

        # Set the environment
        self.environment_variables = self.__environment()

        # Fill in container instructions
        self.__instructions()

    def __instructions(self):
        """Fill in container instructions"""

        self += comment('NVIDIA HPC SDK version {}'.format(self.__version))

        if self.package:
            # Use package from local build context
            self += copy(src=self.package,
                         dest=posixpath.join(self.__wd,
                                             posixpath.basename(self.package)))

        if self.__ospackages:
            self += packages(ospackages=self.__ospackages)

        if self.package or self.__tarball or self.__url:
            # tarball install
            self += shell(commands=self.__commands)
        else:
            # repository install
            if StrictVersion(self.__version) >= StrictVersion('22.9'):
                # signed packages
                self += packages(
                    apt_keys=['https://developer.download.nvidia.com/hpc-sdk/ubuntu/DEB-GPG-KEY-NVIDIA-HPC-SDK'],
                    apt_repositories=['deb [signed-by=/usr/share/keyrings/DEB-GPG-KEY-NVIDIA-HPC-SDK.gpg] https://developer.download.nvidia.com/hpc-sdk/ubuntu/{} /'.format(self.__arch_label)],
                    ospackages=[self.__nvhpc_package],
                    yum_repositories=['https://developer.download.nvidia.com/hpc-sdk/rhel/nvhpc.repo'],
                    _apt_key=False)
            else:
                self += packages(
                    apt_repositories=['deb [trusted=yes] https://developer.download.nvidia.com/hpc-sdk/ubuntu/{} /'.format(self.__arch_label)],
                    ospackages=[self.__nvhpc_package],
                    yum_repositories=['https://developer.download.nvidia.com/hpc-sdk/rhel/nvhpc.repo'])

        if self.__toolchain:
            # Regenerate the localrc using the compilers from the specified
            # toolchain
            compiler_bin = posixpath.join(self.__basepath, 'compilers', 'bin')

            args = ['-x']
            if self.__toolchain.CC:
                args.append('-gcc {}'.format(self.__toolchain.CC))
            if self.__toolchain.CXX:
                args.append('-gpp {}'.format(self.__toolchain.CXX))
            if self.__toolchain.F77:
                args.append('-g77 {}'.format(self.__toolchain.F77))

            self += shell(commands=['{0} {1} {2}'.format(
                posixpath.join(compiler_bin, 'makelocalrc'), compiler_bin,
                ' '.join(args))])

        self += environment(variables=self.environment_step())

    def __cpu_arch(self):
        """Based on the CPU architecture, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_cpu_arch == cpu_arch.AARCH64:
            self.__arch_directory = 'Linux_aarch64'
            if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
                self.__arch_label = 'arm64'
            else:
                self.__arch_label = 'aarch64'
            if StrictVersion(self.__version) < StrictVersion('20.11'):
                self.__cuda_multi = False # CUDA multi packages not available
        elif hpccm.config.g_cpu_arch == cpu_arch.PPC64LE:
            self.__arch_directory = 'Linux_ppc64le'
            if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
                self.__arch_label = 'ppc64el'
            else:
                self.__arch_label = 'ppc64le'
        elif hpccm.config.g_cpu_arch == cpu_arch.X86_64:
            self.__arch_directory = 'Linux_x86_64'
            if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
                self.__arch_label = 'amd64'
            else:
                self.__arch_label = 'x86_64'
        else: # pragma: no cover
            raise RuntimeError('Unknown CPU architecture')

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            version = self.__version.replace('.', '-')
            if self.__cuda_multi:
                self.__nvhpc_package = 'nvhpc-{}-cuda-multi'.format(version)
            else:
                self.__nvhpc_package = 'nvhpc-{}'.format(version)

            if not self.__ospackages:
                if self.package or self.__tarball:
                    self.__ospackages = ['bc', 'debianutils', 'gcc', 'g++',
                                         'gfortran', 'libatomic1', 'libnuma1',
                                         'openssh-client', 'wget']
                else:
                    self.__ospackages = ['ca-certificates', 'gnupg', 'wget']
            self.__runtime_ospackages = ['libatomic1', 'libnuma1',
                                         'openssh-client']
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if self.__cuda_multi:
                self.__nvhpc_package = 'nvhpc-cuda-multi-{}'.format(self.__version)
            else:
                self.__nvhpc_package = 'nvhpc-{}'.format(self.__version)

            if not self.__ospackages:
                if self.package or self.__tarball:
                    self.__ospackages = ['bc', 'gcc', 'gcc-c++',
                                         'gcc-gfortran', 'libatomic',
                                         'openssh-clients', 'numactl-libs',
                                         'wget', 'which']
                else:
                    self.__ospackages = ['ca-certificates']
            self.__runtime_ospackages = ['libatomic', 'numactl-libs',
                                         'openssh-clients']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def __environment(self):
        """Define environment variables"""
        e = {}

        # Development environment
        if self.__extended_environment:
            # Mirror the environment defined by the environment module
            e['CC'] = posixpath.join(self.__basepath, 'compilers', 'bin',
                                     'nvc')
            e['CPP'] = 'cpp'
            e['CXX'] = posixpath.join(self.__basepath, 'compilers', 'bin',
                                      'nvc++')
            e['F77'] = posixpath.join(self.__basepath, 'compilers', 'bin',
                                      'nvfortran')
            e['F90'] = posixpath.join(self.__basepath, 'compilers', 'bin',
                                      'nvfortran')
            e['FC'] = posixpath.join(self.__basepath, 'compilers', 'bin',
                                     'nvfortran')

        cpath = [
            posixpath.join(self.__basepath, 'comm_libs', 'nvshmem', 'include'),
            posixpath.join(self.__basepath, 'comm_libs', 'nccl', 'include'),
            posixpath.join(self.__basepath, 'compilers', 'extras', 'qd',
                           'include', 'qd'),
            posixpath.join(self.__basepath, 'math_libs', 'include')]

        ld_library_path = [
            posixpath.join(self.__basepath, 'comm_libs', 'nvshmem', 'lib'),
            posixpath.join(self.__basepath, 'comm_libs', 'nccl', 'lib'),
            posixpath.join(self.__basepath, 'math_libs', 'lib64'),
            posixpath.join(self.__basepath, 'compilers', 'lib'),
            posixpath.join(self.__basepath, 'cuda', 'lib64')]

        path = [
            posixpath.join(self.__basepath, 'comm_libs', 'nvshmem', 'bin'),
            posixpath.join(self.__basepath, 'comm_libs', 'nccl', 'bin'),
            posixpath.join(self.__basepath, 'profilers', 'bin'),
            posixpath.join(self.__basepath, 'compilers', 'bin'),
            posixpath.join(self.__basepath, 'cuda', 'bin')]

        if self.__mpi:
            cpath.append(
                posixpath.join(self.__basepath, 'comm_libs', 'mpi', 'include'))
            ld_library_path.append(
                posixpath.join(self.__basepath, 'comm_libs', 'mpi', 'lib'))
            path.append(
                posixpath.join(self.__basepath, 'comm_libs', 'mpi', 'bin'))
        elif self.__hpcx:
            # Set environment for HPC-X
            if StrictVersion(self.__version) >= StrictVersion('22.2'):
                hpcx_version = 'latest'
            elif StrictVersion(self.__version) >= StrictVersion('21.11'):
                hpcx_version = 'hpcx-2.10.beta'
            elif StrictVersion(self.__version) >= StrictVersion('21.9'):
                hpcx_version = 'hpcx-2.9.0'
            elif StrictVersion(self.__version) >= StrictVersion('21.7'):
                hpcx_version = 'hpcx-2.8.1'
            elif StrictVersion(self.__version) < StrictVersion('21.5'):
              hpcx_version = 'hpcx-2.7.4'
            hpcx_dir = posixpath.join(self.__basepath, 'comm_libs', 'hpcx',
                                      hpcx_version)

            hpcx_ucx_dir = posixpath.join(hpcx_dir, 'ucx', 'mt')
            #hpcx_ucx_dir = posixpath.join(hpcx_dir, 'ucx')
            hpcx_sharp_dir = posixpath.join(hpcx_dir, 'sharp')
            hpcx_nccl_rdma_sharp_plugin_dir = posixpath.join(
                hpcx_dir, 'nccl_rdma_sharp_plugin')
            hpcx_hcoll_dir = posixpath.join(hpcx_dir, 'hcoll')
            hpcx_mpi_dir = posixpath.join(hpcx_dir, 'ompi')
            hpcx_oshmem_dir = hpcx_mpi_dir

            cpath.append(':'.join([
                posixpath.join(hpcx_hcoll_dir, 'include'),
                posixpath.join(hpcx_mpi_dir, 'include'),
                posixpath.join(hpcx_sharp_dir, 'include'),
                posixpath.join(hpcx_ucx_dir, 'include'),
                '$CPATH']))
            e['HPCX_DIR'] = hpcx_dir
            e['HPCX_HCOLL_DIR'] = hpcx_hcoll_dir
            e['HPCX_MPI_DIR'] = hpcx_mpi_dir
            e['HPCX_NCCL_RDMA_SHARP_PLUGIN_DIR'] = hpcx_nccl_rdma_sharp_plugin_dir
            e['HPCX_OSHMEM_DIR'] = hpcx_oshmem_dir
            e['HPCX_SHARP_DIR'] = hpcx_sharp_dir
            e['HPCX_UCX_DIR'] = hpcx_ucx_dir
            e['LIBRARY_PATH'] = ':'.join([
                posixpath.join(hpcx_hcoll_dir, 'lib'),
                posixpath.join(hpcx_mpi_dir, 'lib'),
                posixpath.join(hpcx_nccl_rdma_sharp_plugin_dir, 'lib'),
                posixpath.join(hpcx_sharp_dir, 'lib'),
                posixpath.join(hpcx_ucx_dir, 'lib'),
                '$LIBRARY_PATH'])
            ld_library_path.append(':'.join([
                posixpath.join(hpcx_hcoll_dir, 'lib'),
                posixpath.join(hpcx_mpi_dir, 'lib'),
                posixpath.join(hpcx_nccl_rdma_sharp_plugin_dir, 'lib'),
                posixpath.join(hpcx_sharp_dir, 'lib'),
                posixpath.join(hpcx_ucx_dir, 'lib'),
                posixpath.join(hpcx_ucx_dir, 'lib', 'ucx'),
                '$LD_LIBRARY_PATH']))
            e['MPI_HOME'] = hpcx_mpi_dir
            e['OMPI_HOME'] = hpcx_mpi_dir
            e['OPAL_PREFIX'] = hpcx_mpi_dir
            e['OSHMEM_HOME'] = hpcx_mpi_dir
            path.append(':'.join([
                posixpath.join(hpcx_hcoll_dir, 'bin'),
                posixpath.join(hpcx_mpi_dir, 'bin'),
                posixpath.join(hpcx_ucx_dir, 'bin'),
                '$PATH']))
            e['PKG_CONFIG_PATH'] = ':'.join([
                posixpath.join(hpcx_hcoll_dir, 'lib', 'pkgconfig'),
                posixpath.join(hpcx_mpi_dir, 'lib', 'pkgconfig'),
                posixpath.join(hpcx_sharp_dir, 'lib', 'pkgconfig'),
                posixpath.join(hpcx_ucx_dir, 'lib', 'pkgconfig'),
                '$PKG_CONFIG_PATH'])
            e['SHMEM_HOME'] = hpcx_mpi_dir

        if cpath:
            e['CPATH'] = '{}:$CPATH'.format(':'.join(cpath))
        e['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(':'.join(
            ld_library_path))
        e['MANPATH'] = '{}:$MANPATH'.format(
            posixpath.join(self.__basepath, 'compilers', 'man'))
        e['PATH'] = '{}:$PATH'.format(':'.join(path))

        return e

    def __get_version(self):
        """Figure out the version information"""

        if self.package:
            # Figure out the version from the package name
            match = re.search(r'nvhpc_\d+_(?P<year>\d\d)0?(?P<month>[1-9][0-9]?)',
                              self.package)
            if (match and match.groupdict()['year'] and
                match.groupdict()['month']):
                self.__version = '{0}.{1}'.format(match.groupdict()['year'],
                                                  match.groupdict()['month'])
                self.__year = '20' + match.groupdict()['year']
            else:
                raise RuntimeError('could not parse version from package name')
        else:
            match = re.search(r'(?P<year>\d\d)\.\d+', self.__version)
            if match and match.groupdict()['year']:
                self.__year = '20' + match.groupdict()['year']

    def __setup_tarball(self):
        """Construct the series of shell commands, i.e., fill in
           self.__commands"""

        # Download / copy package
        if not self.package:
            if self.__url:
                self.url = self.__url
            else:
                baseurl = 'https://developer.download.nvidia.com/hpc-sdk/{0}/nvhpc_{1}_{2}_{3}_cuda_{{}}.tar.gz'.format(
                    self.__version, self.__year,
                    self.__version.replace('.', ''), self.__arch_directory)
                if self.__cuda_multi:
                    self.url = baseurl.format('multi')
                else:
                    self.url = baseurl.format(
                        self.__cuda_version if self.__cuda_version
                        else self.__cuda_version_default)

        self.__commands.append(self.download_step(wd=self.__wd))

        # Set installer flags
        flags = {'NVHPC_ACCEPT_EULA': 'accept',
                 'NVHPC_INSTALL_DIR': self.__prefix,
                 'NVHPC_SILENT': 'true'}
        if self.__cuda_version:
            flags['NVHPC_DEFAULT_CUDA'] = self.__cuda_version
        if self.__stdpar_cudacc:
            flags['NVHPC_STDPAR_CUDACC'] = self.__stdpar_cudacc
        if not self.__eula:
            # This will fail when building the container
            logging.warning('NVIDIA HPC SDK EULA was not accepted')
            flags['NVHPC_ACCEPT_EULA'] = 'decline'
            flags['NVHPC_SILENT'] = 'false'
        flag_string = ' '.join('{0}={1}'.format(key, val)
                               for key, val in sorted(flags.items()))

        # Install
        self.__commands.append('cd {0} && {1} ./install'.format(
            self.src_directory, flag_string))

        # Cleanup
        remove = [self.src_directory]
        if self.url:
            remove.append(posixpath.join(self.__wd,
                                         posixpath.basename(self.url)))
        elif self.package:
            remove.append(posixpath.join(self.__wd,
                                         posixpath.basename(self.package)))
        self.__commands.append(self.cleanup_step(items=remove))

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        n = nvhpc(redist=[...], ...)
        Stage0 += n
        Stage1 += n.runtime()
        ```
        """
        if self.__redist:
            self.rt += comment('NVIDIA HPC SDK')

            if self.__runtime_ospackages:
                self.rt += packages(ospackages=self.__runtime_ospackages)

            redistpath = posixpath.join(self.__prefix,
                                        self.__arch_directory,
                                        self.__version, 'REDIST')

            libdirs = {}

            for r in self.__redist:
                src = posixpath.join(redistpath, r)

                if '*' in posixpath.basename(r):
                    # When using COPY with more than one source file,
                    # the destination must be a directory and end with
                    # a /
                    dest = posixpath.join(posixpath.dirname(redistpath),
                                          posixpath.dirname(r)) + '/'
                else:
                    dest = posixpath.join(posixpath.dirname(redistpath), r)

                self.rt += copy(_from=_from, src=src, dest=dest)

                # If the redist path looks like a library directory,
                # add it to LD_LIBRARY_PATH
                if '/lib' in posixpath.dirname(r):
                    libdirs[posixpath.join(posixpath.dirname(redistpath),
                                           posixpath.dirname(r))] = True

            if self.__redist and self.__mpi:
                mpipath = posixpath.join(self.__basepath, 'comm_libs', 'mpi')
                self.rt += copy(_from=_from, src=mpipath, dest=mpipath)
                libdirs[posixpath.join(mpipath, 'lib')] = True
                self.runtime_environment_variables['PATH'] = '{}:$PATH'.format(
                    posixpath.join(mpipath, 'bin'))

            if libdirs:
                liblist = sorted(libdirs.keys())
                liblist.append('$LD_LIBRARY_PATH')
                
                self.runtime_environment_variables['LD_LIBRARY_PATH'] = ':'.join(liblist)
                self.rt += environment(
                    variables=self.runtime_environment_variables)

        return str(self.rt)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the mlnx_ofed module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import aarch64, centos, centos8, docker, ppc64le, ubuntu, ubuntu18, ubuntu20, x86_64

from hpccm.building_blocks.mlnx_ofed import mlnx_ofed

class Test_mlnx_ofed(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @x86_64
    @ubuntu
    @docker
    def test_defaults_ubuntu(self):
        """Default mlnx_ofed building block"""
        mofed = mlnx_ofed()
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 5.2-2.2.0.0
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox | apt-key add - && \
    mkdir -p /etc/apt/sources.list.d && wget -q -nc --no-check-certificate -P /etc/apt/sources.list.d https://linux.mellanox.com/public/repo/mlnx_ofed/5.2-2.2.0.0/ubuntu16.04/mellanox_mlnx_ofed.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ibverbs-providers \
        ibverbs-utils \
        libibmad-dev \
        libibmad5 \
        libibumad-dev \
        libibumad3 \
        libibverbs-dev \
        libibverbs1 \
        librdmacm-dev \
        librdmacm1 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @ubuntu18
    @docker
    def test_defaults_ubuntu18(self):
        """Default mlnx_ofed building block"""
        mofed = mlnx_ofed()
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 5.2-2.2.0.0
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox | apt-key add - && \
    mkdir -p /etc/apt/sources.list.d && wget -q -nc --no-check-certificate -P /etc/apt/sources.list.d https://linux.mellanox.com/public/repo/mlnx_ofed/5.2-2.2.0.0/ubuntu18.04/mellanox_mlnx_ofed.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ibverbs-providers \
        ibverbs-utils \
        libibmad-dev \
        libibmad5 \
        libibumad-dev \
        libibumad3 \
        libibverbs-dev \
        libibverbs1 \
        librdmacm-dev \
        librdmacm1 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @ubuntu20
    @docker
    def test_defaults_ubuntu20(self):
        """Default mlnx_ofed building block"""
        mofed = mlnx_ofed()
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 5.2-2.2.0.0
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox | apt-key add - && \
    mkdir -p /etc/apt/sources.list.d && wget -q -nc --no-check-certificate -P /etc/apt/sources.list.d https://linux.mellanox.com/public/repo/mlnx_ofed/5.2-2.2.0.0/ubuntu20.04/mellanox_mlnx_ofed.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ibverbs-providers \
        ibverbs-utils \
        libibmad-dev \
        libibmad5 \
        libibumad-dev \
        libibumad3 \
        libibverbs-dev \
        libibverbs1 \
        librdmacm-dev \
        librdmacm1 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @centos
    @docker
    def test_defaults_centos(self):
        """Default mlnx_ofed building block"""
        mofed = mlnx_ofed()
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 5.2-2.2.0.0
RUN yum install -y \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/cache/yum/*
RUN rpm --import https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox && \
    yum install -y yum-utils && \
    yum-config-manager --add-repo https://linux.mellanox.com/public/repo/mlnx_ofed/5.2-2.2.0.0/rhel7.2/mellanox_mlnx_ofed.repo && \
    yum install -y \
        libibumad \
        libibverbs \
        libibverbs-utils \
        librdmacm \
        rdma-core \
        rdma-core-devel && \
    rm -rf /var/cache/yum/*''')

    @x86_64
    @centos8
    @docker
    def test_defaults_centos8(self):
        """Default mlnx_ofed building block"""
        mofed = mlnx_ofed()
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 5.2-2.2.0.0
RUN yum install -y \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/cache/yum/*
RUN rpm --import https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox && \
    yum install -y dnf-utils && \
    yum-config-manager --add-repo https://linux.mellanox.com/public/repo/mlnx_ofed/5.2-2.2.0.0/rhel8.0/mellanox_mlnx_ofed.repo && \
    yum install -y \
        libibumad \
        libibverbs \
        libibverbs-utils \
        librdmacm \
        rdma-core \
        rdma-core-devel && \
    rm -rf /var/cache/yum/*''')

    @x86_64
    @ubuntu
    @docker
    def test_prefix_ubuntu(self):
        """Prefix option"""
        mofed = mlnx_ofed(prefix='/opt/ofed', version='4.6-1.0.1.1')
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 4.6-1.0.1.1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ca-certificates \
        gnupg \
        libnl-3-200 \
        libnl-route-3-200 \
        libnuma1 \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox | apt-key add - && \
    mkdir -p /etc/apt/sources.list.d && wget -q -nc --no-check-certificate -P /etc/apt/sources.list.d https://linux.mellanox.com/public/repo/mlnx_ofed/4.6-1.0.1.1/ubuntu16.04/mellanox_mlnx_ofed.list && \
    apt-get update -y && \
    mkdir -m 777 -p /var/tmp/packages_download && cd /var/tmp/packages_download && \
    DEBIAN_FRONTEND=noninteractive apt-get download -y --no-install-recommends \
        ibverbs-utils \
        libibmad \
        libibmad-devel \
        libibumad \
        libibumad-devel \
        libibverbs-dev \
        libibverbs1 \
        libmlx4-1 \
        libmlx4-dev \
        libmlx5-1 \
        libmlx5-dev \
        librdmacm-dev \
        librdmacm1 && \
    mkdir -p /opt/ofed && \
    find /var/tmp/packages_download -regextype posix-extended -type f -regex "/var/tmp/packages_download/(ibverbs-utils|libibmad|libibmad-devel|libibumad|libibumad-devel|libibverbs-dev|libibverbs1|libmlx4-1|libmlx4-dev|libmlx5-1|libmlx5-dev|librdmacm-dev|librdmacm1).*deb" -exec dpkg --extract {} /opt/ofed \; && \
    rm -rf /var/tmp/packages_download && \
    rm -f /etc/apt/sources.list.d/mellanox_mlnx_ofed.list && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /etc/libibverbs.d''')

    @x86_64
    @centos
    @docker
    def test_prefix_centos(self):
        """Prefix option"""
        mofed = mlnx_ofed(prefix='/opt/ofed', version='4.6-1.0.1.1')
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 4.6-1.0.1.1
RUN yum install -y \
        ca-certificates \
        gnupg \
        libnl \
        libnl3 \
        numactl-libs \
        wget && \
    rm -rf /var/cache/yum/*
RUN rpm --import https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox && \
    yum install -y yum-utils && \
    yum-config-manager --add-repo https://linux.mellanox.com/public/repo/mlnx_ofed/4.6-1.0.1.1/rhel7.2/mellanox_mlnx_ofed.repo && \
    yum install -y yum-utils && \
    mkdir -p /var/tmp/packages_download && \
    yumdownloader --destdir=/var/tmp/packages_download -x \*i?86 --archlist=x86_64 \
        libibmad \
        libibmad-devel \
        libibumad \
        libibumad-devel \
        libibverbs \
        libibverbs-devel \
        libibverbs-utils \
        libmlx4 \
        libmlx4-devel \
        libmlx5 \
        libmlx5-devel \
        librdmacm \
        librdmacm-devel && \
    mkdir -p /opt/ofed && cd /opt/ofed && \
    find /var/tmp/packages_download -regextype posix-extended -type f -regex "/var/tmp/packages_download/(libibmad|libibmad-devel|libibumad|libibumad-devel|libibverbs|libibverbs-devel|libibverbs-utils|libmlx4|libmlx4-devel|libmlx5|libmlx5-devel|librdmacm|librdmacm-devel).*rpm" -exec sh -c "rpm2cpio {} | cpio -idm" \; && \
    rm -rf /var/tmp/packages_download && \
    rm -rf /var/cache/yum/*
RUN mkdir -p /etc/libibverbs.d''')

    @aarch64
    @centos
    @docker
    def test_aarch64_centos(self):
        """aarch64"""
        mofed = mlnx_ofed(version='4.6-1.0.1.1')
        self.assertEqual(str(mofed),
r'''# Mellanox OFED version 4.6-1.0.1.1
RUN yum install -y \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/cache/yum/*
RUN rpm --import https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox && \
    yum install -y yum-utils && \
    yum-config-manager --add-repo https://linux.mellanox.com/public/repo/mlnx_ofed/4.6-1.0.1.1/rhel7.6alternate/mellanox_mlnx_ofed.repo && \
    yum install -y \
        libibmad \
        libibmad-devel \
        libibumad \
        libibumad-devel \
        libibverbs \
        libibverbs-devel \
        libibverbs-utils \
        libmlx4 \
        libmlx4-devel \
        libmlx5 \
        libmlx5-devel \
        librdmacm \
        librdmacm-devel && \
    rm -rf /var/cache/yum/*''')

    @x86_64
    @ubuntu
    @docker
    def test_runtime(self):
        """Runtime"""
        mofed = mlnx_ofed(version='5.0-2.1.8.0')
        r = mofed.runtime()
        self.assertEqual(r,
r'''# Mellanox OFED version 5.0-2.1.8.0
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://www.mellanox.com/downloads/ofed/RPM-GPG-KEY-Mellanox | apt-key add - && \
    mkdir -p /etc/apt/sources.list.d && wget -q -nc --no-check-certificate -P /etc/apt/sources.list.d https://linux.mellanox.com/public/repo/mlnx_ofed/5.0-2.1.8.0/ubuntu16.04/mellanox_mlnx_ofed.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        ibverbs-providers \
        ibverbs-utils \
        libibmad-dev \
        libibmad5 \
        libibumad-dev \
        libibumad3 \
        libibverbs-dev \
        libibverbs1 \
        librdmacm-dev \
        librdmacm1 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @ubuntu
    @docker
    def test_prefix_runtime(self):
        """Prefix runtime"""
        mofed = mlnx_ofed(prefix='/opt/ofed', version='5.0-2.1.8.0')
        r = mofed.runtime()
        self.assertEqual(r,
r'''# Mellanox OFED version 5.0-2.1.8.0
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        libnl-3-200 \
        libnl-route-3-200 \
        libnuma1 && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /etc/libibverbs.d
COPY --from=0 /opt/ofed /opt/ofed''')



"""
MPI Bandwidth

Contents:
  CentOS 7
  GNU compilers (upstream)
  Mellanox OFED
  OpenMPI
  PMI2 (SLURM)
  UCX

Building:
  1. Docker to Singularity
     $ hpccm --recipe mpi_bandwidth.py > Dockerfile
     $ sudo docker build -t mpi_bw -f Dockerfile .
     $ singularity build mpi_bw.sif docker-daemon://mpi_bw:latest

  2. Singularity
     $ hpccm --recipe mpi_bandwidth.py --format singularity --singularity-version=3.2 > Singularity.def
     $ sudo singularity build mpi_bw.sif Singularity.def

Running with Singularity:
  1. Using a compatible host MPI runtime
     $ mpirun -n 2 singularity run mpi_bw.sif mpi_bandwidth

  2. Using the MPI runtime inside the container
     $ singularity run mpi_bw.sif mpirun -n 2 -H node1:1,node2:1 --launch-agent "singularity exec \$SINGULARITY_CONTAINER orted" mpi_bandwidth

  3. Using SLURM srun
     $ srun -n 2 --mpi=pmi2 singularity run mpi_bw.sif mpi_bandwidth
"""

Stage0 += comment(__doc__, reformat=False)

# CentOS base image
Stage0 += baseimage(image='centos:7', _as='build')

# GNU compilers
Stage0 += gnu(fortran=False)

# Mellanox OFED
Stage0 += mlnx_ofed()

# UCX
Stage0 += ucx(cuda=False)

# PMI2
Stage0 += slurm_pmi2()

# OpenMPI (use UCX instead of IB directly)
Stage0 += openmpi(cuda=False, infiniband=False, pmi='/usr/local/slurm-pmi2',
                  ucx='/usr/local/ucx')

# MPI Bandwidth
Stage0 += shell(commands=[
    'wget -q -nc --no-check-certificate -P /var/tmp https://hpc-tutorials.llnl.gov/mpi/examples/mpi_bandwidth.c',
    'mpicc -o /usr/local/bin/mpi_bandwidth /var/tmp/mpi_bandwidth.c'])

### Runtime distributable stage
Stage1 += baseimage(image='centos:7')
Stage1 += Stage0.runtime()
Stage1 += copy(_from='build', src='/usr/local/bin/mpi_bandwidth',
               dest='/usr/local/bin/mpi_bandwidth')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the mvapich2 module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import centos, docker, ubuntu

from hpccm.building_blocks.mvapich2 import mvapich2
from hpccm.toolchain import toolchain

class Test_mvapich2(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @ubuntu
    @docker
    def test_defaults_ubuntu(self):
        """Default mvapich2 building block"""
        mv2 = mvapich2()
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3.4
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        byacc \
        file \
        flex \
        make \
        openssh-client \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3.4.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3.4.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3.4 && \
    ln -s /usr/local/cuda/lib64/stubs/libnvidia-ml.so /usr/local/cuda/lib64/stubs/libnvidia-ml.so.1 && \
    ln -s /usr/local/cuda/lib64/stubs/libcuda.so /usr/local/cuda/lib64/stubs/libcuda.so.1 && \
    cd /var/tmp/mvapich2-2.3.4 &&  LD_LIBRARY_PATH='/usr/local/cuda/lib64/stubs:$LD_LIBRARY_PATH' ./configure --prefix=/usr/local/mvapich2 --disable-mcast --enable-cuda --with-cuda=/usr/local/cuda && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mvapich2-2.3.4 /var/tmp/mvapich2-2.3.4.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH \
    PROFILE_POSTLIB="-L/usr/local/cuda/lib64/stubs -lnvidia-ml -lcuda"''')

    @ubuntu
    @docker
    def test_pgi_cuda(self):
        """mvapich2 with pgi and cuda"""
        tc = toolchain()
        tc.CC = 'pgcc'
        tc.CXX = 'pgc++'
        tc.F77 = 'pgfortran'
        tc.FC = 'pgfortran'
        mv2 = mvapich2(toolchain=tc, cuda=True, version='2.3.3')
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3.3
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        byacc \
        file \
        flex \
        make \
        openssh-client \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3.3.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3.3.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3.3 && \
    ln -s /usr/local/cuda/lib64/stubs/libnvidia-ml.so /usr/local/cuda/lib64/stubs/libnvidia-ml.so.1 && \
    ln -s /usr/local/cuda/lib64/stubs/libcuda.so /usr/local/cuda/lib64/stubs/libcuda.so.1 && \
    cd /var/tmp/mvapich2-2.3.3 &&  CC=pgcc CFLAGS=-ta=tesla:nordc CPPFLAGS='-D__x86_64 -D__align__\(n\)=__attribute__\(\(aligned\(n\)\)\) -D__location__\(a\)=__annotate__\(a\) -DCUDARTAPI=' CXX=pgc++ F77=pgfortran FC=pgfortran LD_LIBRARY_PATH='/usr/local/cuda/lib64/stubs:$LD_LIBRARY_PATH' ./configure --prefix=/usr/local/mvapich2 --disable-mcast --enable-cuda=basic --with-cuda=/usr/local/cuda --enable-fast=O1 && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mvapich2-2.3.3 /var/tmp/mvapich2-2.3.3.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH \
    PROFILE_POSTLIB="-L/usr/local/cuda/lib64/stubs -lnvidia-ml -lcuda"''')

    @ubuntu
    @docker
    def test_nvhpc(self):
        """mvapich2 with NVIDIA HPC SDK"""
        tc = toolchain()
        tc.CC = 'nvc'
        tc.CXX = 'nvc++'
        tc.F77 = 'nvfortran'
        tc.FC = 'nvfortran'
        mv2 = mvapich2(toolchain=tc, cuda=False, version='2.3.3')
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3.3
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        byacc \
        file \
        flex \
        make \
        openssh-client \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3.3.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3.3.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3.3 &&  CC=nvc CFLAGS='-fpic -DPIC' CXX=nvc++ F77=nvfortran FC=nvfortran FCFLAGS='-fpic -DPIC' FFLAGS='-fpic -DPIC' ./configure --prefix=/usr/local/mvapich2 --disable-cuda --disable-mcast ac_cv_c_compiler_gnu=no && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mvapich2-2.3.3 /var/tmp/mvapich2-2.3.3.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH''')

    @ubuntu
    @docker
    def test_gpu_arch(self):
        """mvapich2 GPU architecture"""
        mv2 = mvapich2(version='2.3b', gpu_arch='sm_60')
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3b
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        byacc \
        file \
        flex \
        make \
        openssh-client \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3b.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3b.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3b && \
    ln -s /usr/local/cuda/lib64/stubs/libnvidia-ml.so /usr/local/cuda/lib64/stubs/libnvidia-ml.so.1 && \
    ln -s /usr/local/cuda/lib64/stubs/libcuda.so /usr/local/cuda/lib64/stubs/libcuda.so.1 && \
    sed -i -e 's/-arch sm_20/-arch sm_60/g' Makefile.in && \
    cd /var/tmp/mvapich2-2.3b &&  LD_LIBRARY_PATH='/usr/local/cuda/lib64/stubs:$LD_LIBRARY_PATH' ./configure --prefix=/usr/local/mvapich2 --disable-mcast --enable-cuda --with-cuda=/usr/local/cuda && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mvapich2-2.3b /var/tmp/mvapich2-2.3b.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH \
    PROFILE_POSTLIB="-L/usr/local/cuda/lib64/stubs -lnvidia-ml -lcuda"''')

    @ubuntu
    @docker
    def test_nocuda(self):
        """Disable CUDA"""
        mv2 = mvapich2(cuda=False, version='2.3.3')
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3.3
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        byacc \
        file \
        flex \
        make \
        openssh-client \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3.3.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3.3.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3.3 &&   ./configure --prefix=/usr/local/mvapich2 --disable-cuda --disable-mcast && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mvapich2-2.3.3 /var/tmp/mvapich2-2.3.3.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH''')

    @centos
    @docker
    def test_defaults_centos(self):
        """Default mvapich2 building block"""
        mv2 = mvapich2()
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3.4
RUN yum install -y \
        byacc \
        file \
        flex \
        make \
        openssh-clients \
        wget && \
    rm -rf /var/cache/yum/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3.4.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3.4.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3.4 && \
    ln -s /usr/local/cuda/lib64/stubs/libnvidia-ml.so /usr/local/cuda/lib64/stubs/libnvidia-ml.so.1 && \
    ln -s /usr/local/cuda/lib64/stubs/libcuda.so /usr/local/cuda/lib64/stubs/libcuda.so.1 && \
    cd /var/tmp/mvapich2-2.3.4 &&  LD_LIBRARY_PATH='/usr/local/cuda/lib64/stubs:$LD_LIBRARY_PATH' ./configure --prefix=/usr/local/mvapich2 --disable-mcast --enable-cuda --with-cuda=/usr/local/cuda && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mvapich2-2.3.4 /var/tmp/mvapich2-2.3.4.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH \
    PROFILE_POSTLIB="-L/usr/local/cuda/lib64/stubs -lnvidia-ml -lcuda"''')

    @ubuntu
    @docker
    def test_ldconfig(self):
        """ldconfig option"""
        mv2 = mvapich2(ldconfig=True, version='2.3')
        self.assertEqual(str(mv2),
r'''# MVAPICH2 version 2.3
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        byacc \
        file \
        flex \
        make \
        openssh-client \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp http://mvapich.cse.ohio-state.edu/download/mvapich/mv2/mvapich2-2.3.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mvapich2-2.3.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mvapich2-2.3 && \
    ln -s /usr/local/cuda/lib64/stubs/libnvidia-ml.so /usr/local/cuda/lib64/stubs/libnvidia-ml.so.1 && \
    ln -s /usr/local/cuda/lib64/stubs/libcuda.so /usr/local/cuda/lib64/stubs/libcuda.so.1 && \
    cd /var/tmp/mvapich2-2.3 &&  LD_LIBRARY_PATH='/usr/local/cuda/lib64/stubs:$LD_LIBRARY_PATH' ./configure --prefix=/usr/local/mvapich2 --disable-mcast --enable-cuda --with-cuda=/usr/local/cuda && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    echo "/usr/local/mvapich2/lib" >> /etc/ld.so.conf.d/hpccm.conf && ldconfig && \
    rm -rf /var/tmp/mvapich2-2.3 /var/tmp/mvapich2-2.3.tar.gz
ENV PATH=/usr/local/mvapich2/bin:$PATH \
    PROFILE_POSTLIB="-L/usr/local/cuda/lib64/stubs -lnvidia-ml -lcuda"''')

    @ubuntu
    @docker
    def test_runtime(self):
        """Runtime"""
        mv2 = mvapich2()
        r = mv2.runtime()
        self.assertEqual(r,
r'''# MVAPICH2
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        openssh-client && \
    rm -rf /var/lib/apt/lists/*
COPY --from=0 /usr/local/mvapich2 /usr/local/mvapich2
ENV LD_LIBRARY_PATH=/usr/local/mvapich2/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mvapich2/bin:$PATH''')

    def test_toolchain(self):
        """Toolchain"""
        mv2 = mvapich2()
        tc = mv2.toolchain
        self.assertEqual(tc.CC, 'mpicc')
        self.assertEqual(tc.CXX, 'mpicxx')
        self.assertEqual(tc.FC, 'mpifort')
        self.assertEqual(tc.F77, 'mpif77')
        self.assertEqual(tc.F90, 'mpif90')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Container stage"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import

import hpccm.config

from hpccm.common import container_type
from hpccm.primitives.baseimage import baseimage

class Stage(object):
    """Class for container stages.

    Docker may have one or more stages,
       Singularity will always have a single stage.

    # Parameters

    name: Name to use when refering to the stage (Docker specific).
    The default is an empty string.

    separator: Separator to insert between stages.  The default is
    '\\n\\n'.

    """

    def __init__(self, **kwargs):
        """Initialize stage"""

        self.__layers = []
        self.name = kwargs.get('name', '')
        self.__separator = kwargs.get('separator', '\n\n')

    def __iadd__(self, layer):
        """Add the layer to the stage.  Allows "+=" syntax."""

        # The name of the stage should reflect the name the user
        # provided in the baseimage primitive (via the _as parameter).
        # This violates the encapsulation of the baseimage primitive.
        if layer.__class__.__name__ == 'baseimage' and not self.name:
          self.name = layer._baseimage__as

        if isinstance(layer, list):
            self.__layers.extend(layer)
        else:
            self.__layers.append(layer)
        return self

    def __len__(self):
        """Return number of layers"""
        return len(self.__layers)

    def __str__(self):
        """String representation of the stage"""
        return self.__separator.join(str(x) for x in self.__layers)

    def baseimage(self, image, _distro=''):
        """Insert the baseimage as the first layer

        # Arguments

        image (string): The image identifier to use as the base image.
        The value is passed to the `baseimage` primitive.

        _distro: The underlying Linux distribution of the base image.
        The value is passed to the `baseimage` primitive.
        """
        if image:
            self.__layers.insert(0, baseimage(image=image, _as=self.name,
                                              _distro=_distro))

    def runtime(self, _from=None, exclude=[]):
        """Generate the set of instructions to install the runtime specific
        components from a previous stage.

        This method invokes the runtime() method for every layer in
        the stage.  If a layer does not have a runtime() method, then
        it is skipped.

        # Arguments

        _from: The name of the stage from which to copy the runtime.
        The default is `0`.

        exclude: List of building blocks to exclude when generating
        the runtime. The default is an empty list.

        # Examples
        ```python
        Stage0 += baseimage(image='nvidia/cuda:9.0-devel')
        Stage0 += gnu()
        Stage0 += boost()
        Stage0 += ofed()
        Stage0 += openmpi()
        ...
        Stage1 += baseimage(image='nvidia/cuda:9.0-base')
        Stage1 += Stage0.runtime(exclude=['boost'])
        ```

        """

        # If the name of the stage is not explicitly specified, use
        # the name of the Stage if available, otherwise 0 (Docker's
        # default)
        if not _from and self.name:
            _from = self.name
        elif not _from:
            if hpccm.config.g_ctype == container_type.SINGULARITY:
                logging.warning('Multi-stage Singularity containers require a named first stage')
            _from = '0'

        instructions = []
        for layer in self.__layers:
            runtime = getattr(layer, 'runtime', None)
            if callable(runtime) and layer.__class__.__name__ not in exclude:
                inst = layer.runtime(_from=_from)
                if inst:
                    instructions.append(inst)

        return self.__separator.join(instructions)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""CGNS building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import posixpath
import re
from copy import copy as _copy

import hpccm.config

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_autotools import generic_autotools
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment
from hpccm.toolchain import toolchain

class cgns(bb_base):
    """The `cgns` building block downloads and installs the
    [CGNS](https://cgns.github.io/index.html) component.

    The [HDF5](#hdf5) building block should be installed prior to this
    building block.

    # Parameters

    annotate: Boolean flag to specify whether to include annotations
    (labels).  The default is False.

    check: Boolean flag to specify whether the test cases should be
    run.  The default is False.

    configure_opts: List of options to pass to `configure`.  The
    default value is `--with-hdf5=/usr/local/hdf5` and `--with-zlib`.

    disable_FEATURE: Flags to control disabling features when
    configuring.  For instance, `disable_foo=True` maps to
    `--disable-foo`.  Underscores in the parameter name are converted
    to dashes.

    enable_FEATURE[=ARG]: Flags to control enabling features when
    configuring.  For instance, `enable_foo=True` maps to
    `--enable-foo` and `enable_foo='yes'` maps to `--enable-foo=yes`.
    Underscores in the parameter name are converted to dashes.

    prefix: The top level install location.  The default value is
    `/usr/local/cgns`.

    ospackages: List of OS packages to install prior to configuring
    and building.  For Ubuntu, the default values are `file`, `make`,
    `wget`, and `zlib1g-dev`.  For RHEL-based Linux distributions the
    default values are `bzip2`, `file`, `make`, `wget` and
    `zlib-devel`.

    toolchain: The toolchain object.  This should be used if
    non-default compilers or other toolchain options are needed.  The
    default is empty.

    version: The version of CGNS source to download.  The default
    value is `4.1.2`.

    with_PACKAGE[=ARG]: Flags to control optional packages when
    configuring.  For instance, `with_foo=True` maps to `--with-foo`
    and `with_foo='/usr/local/foo'` maps to
    `--with-foo=/usr/local/foo`.  Underscores in the parameter name
    are converted to dashes.

    without_PACKAGE: Flags to control optional packages when
    configuring.  For instance `without_foo=True` maps to
    `--without-foo`.  Underscores in the parameter name are converted
    to dashes.

    # Examples

    ```python
    cgns(prefix='/opt/cgns/3.3.1', version='3.3.1')
    ```
    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(cgns, self).__init__(**kwargs)

        self.__baseurl = kwargs.pop('baseurl', 'https://github.com/CGNS/CGNS/archive')
        self.__check = kwargs.pop('check', False)
        self.__configure_opts = kwargs.pop('configure_opts',
                                           ['--with-hdf5=/usr/local/hdf5',
                                            '--with-zlib'])
        self.__ospackages = kwargs.pop('ospackages', [])
        self.__prefix = kwargs.pop('prefix', '/usr/local/cgns')
        self.__toolchain = kwargs.pop('toolchain', toolchain())
        self.__version = kwargs.pop('version', '4.1.2')

        # Set the configuration options
        self.__configure()

        # Set the Linux distribution specific parameters
        self.__distro()

        # Setup build configuration
        self.__bb = generic_autotools(
            annotations={'version': self.__version},
            base_annotation=self.__class__.__name__,
            check=self.__check,
            comment=False,
            configure_opts=self.__configure_opts,
            directory=posixpath.join('CGNS-{}'.format(self.__version), 'src'),
            prefix=self.__prefix,
            toolchain=self.__toolchain,
            url='{0}/v{1}.tar.gz'.format(self.__baseurl, self.__version),
            **kwargs)

        # Container instructions
        self += comment('CGNS version {}'.format(self.__version))
        self += packages(ospackages=self.__ospackages)
        self += self.__bb

    def __configure(self):
        """Setup configure options based on user parameters"""

        # Create a copy of the toolchain so that it can be modified
        # without impacting the original.
        self.__toolchain = _copy(self.__toolchain)

        # See https://cgns.github.io/download.html, Known Bugs
        if not self.__toolchain.LIBS:
            self.__toolchain.LIBS = '-Wl,--no-as-needed -ldl'
        if not self.__toolchain.FLIBS:
            self.__toolchain.FLIBS = '-Wl,--no-as-needed -ldl'
        # See https://cgnsorg.atlassian.net/browse/CGNS-40
        if (not self.__toolchain.FFLAGS and self.__toolchain.FC and
            re.match('.*pgf.*', self.__toolchain.FC)):
            self.__toolchain.FFLAGS = '-Mx,125,0x200'

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['file', 'make', 'wget', 'zlib1g-dev']
            self.__runtime_ospackages = ['zlib1g']
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['bzip2', 'file', 'make', 'wget',
                                     'zlib-devel']
                if self.__check:
                    self.__ospackages.append('diffutils')
            self.__runtime_ospackages = ['zlib']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Example

        ```python
        c = cgns(...)
        Stage0 += c
        Stage1 += c.runtime()
        ```
        """
        self.rt += comment('CGNS')
        self.rt += packages(ospackages=self.__runtime_ospackages)
        self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the scif module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import os
import tempfile
import unittest

from helpers import centos, docker, singularity

from hpccm.building_blocks.python import python
from hpccm.building_blocks.scif import scif

from hpccm.primitives.copy import copy
from hpccm.primitives.comment import comment
from hpccm.primitives.environment import environment
from hpccm.primitives.label import label
from hpccm.primitives.runscript import runscript
from hpccm.primitives.shell import shell

class Test_scif(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @docker
    def test_noname_docker(self):
        """Default scif building block, no name"""
        with self.assertRaises(RuntimeError):
            scif()

    @docker
    def test_defaults_docker(self):
        """Default scif building block"""
        # The scif module generates a SCI-F recipe file.  Use a
        # temporary file location for it rather than polluting the
        # test environment.  This is slightly unrealistic since the
        # Docker build would fail since the file is outside the Docker
        # build environment, but go with it.
        scif_file = tempfile.NamedTemporaryFile(delete=False, suffix='.scif')
        s = scif(name='foo', file=scif_file.name)
        self.assertEqual(str(s),
r'''# SCI-F "foo"
COPY {0} /scif/recipes/{1}
RUN scif install /scif/recipes/{1}'''.format(scif_file.name,
                                             os.path.basename(scif_file.name)))
        try:
            os.unlink(scif_file.name)
        except WindowsError:
            # WindowsError: [Error 32] The process cannot access the file
            # because it is being used by another process
            pass

    @docker
    def test_arguments_docker(self):
        """Default scif building block with _arguments"""
        scif_file = tempfile.NamedTemporaryFile(delete=False, suffix='.scif')
        s = scif(_arguments='--mount=type=bind,target=/scif/apps/foo/src',
                 file=scif_file.name,
                 name='foo')
        self.assertEqual(str(s),
r'''# SCI-F "foo"
COPY {0} /scif/recipes/{1}
RUN --mount=type=bind,target=/scif/apps/foo/src scif install /scif/recipes/{1}'''.format(scif_file.name,
                                             os.path.basename(scif_file.name)))
        try:
            os.unlink(scif_file.name)
        except WindowsError:
            # WindowsError: [Error 32] The process cannot access the file
            # because it is being used by another process
            pass

    @singularity
    def test_defaults_singularity(self):
        """Default scif building block"""
        s = scif(name='foo')
        self.assertEqual(str(s), '')

    @singularity
    def test_allsections_singularity(self):
        """One of each SCI-f section type"""
        s = scif(name='foo')
        s += copy(src='file', dest='/src/file')
        s += comment('My app')
        s += environment(variables={'ONE': '1'})
        s += label(metadata={'A': 'B'})
        s += runscript(commands=['default_program'])
        s += shell(commands=['build_cmds'])
        s += shell(commands=['test_program'], _test=True)
        self.assertEqual(str(s),
r'''%appenv foo
    export ONE=1
%appfiles foo
    file /src/file
%apphelp foo
My app
%appinstall foo
    build_cmds
%applabels foo
    A B
%apprun foo
    exec default_program "$@"
%apptest foo
    test_program''')

    @docker
    def test_allsections_docker(self):
        """One of each SCI-f section type"""
        # See comment in the test_defaults_docker test case
        scif_file = tempfile.NamedTemporaryFile(delete=False, suffix='.scif')
        s = scif(name='foo', file=scif_file.name)
        s += copy(src='file', dest='/src/file')
        s += comment('My app')
        s += environment(variables={'ONE': '1'})
        s += label(metadata={'A': 'B'})
        s += runscript(commands=['default_program'])
        s += shell(commands=['build_cmds'])
        s += shell(commands=['test_program'], _test=True)

        str(s) # Force writing the SCI-F recipe file

        # slurp file content
        with open(scif_file.name) as f:
            content = f.read()
        try:
            os.unlink(scif_file.name)
        except WindowsError:
            # WindowsError: [Error 32] The process cannot access the file
            # because it is being used by another process
            pass

        self.assertEqual(content,
r'''%appenv foo
    export ONE=1

%appfiles foo
    file /src/file

%apphelp foo
My app

%appinstall foo
    build_cmds

%applabels foo
    A B

%apprun foo
    exec default_program "$@"

%apptest foo
    test_program''')

    @centos
    @singularity
    def test_building_block_singularity(self):
        """test building block"""
        s = scif(name='foo')
        s += [python()] # list not required, but tests a code branch
        self.assertEqual(str(s),
r'''%apphelp foo
Python
%appinstall foo
    yum install -y \
        python2 \
        python3
    rm -rf /var/cache/yum/*''')

    @docker
    def test_runtime(self):
        """Runtime"""
        s = scif(name='foo')
        r = s.runtime()
        self.assertEqual(r, 'COPY --from=0 /scif /scif')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

__all__ = ['amgx',
           'apt_get',
           'arm_allinea_studio',
           'boost',
           'catalyst',
           'cgns',
           'charm',
           'cmake',
           'conda',
           'fftw',
           'gdrcopy',
           'generic_autotools',
           'generic_build',
           'generic_cmake',
           'gnu',
           'hdf5',
           'hpcx',
           'intel_mpi',
           'intel_psxe',
           'intel_psxe_runtime',
           'julia',
           'knem',
           'kokkos',
           'libsim',
           'llvm',
           'magma',
           'mkl',
           'mlnx_ofed',
           'mpich',
           'multi_ofed',
           'mvapich2_gdr',
           'mvapich2',
           'nccl',
           'netcdf',
           'nsight_compute',
           'nsight_systems',
           'nvhpc',
           'nvshmem',
           'ofed',
           'openblas',
           'openmpi',
           'packages',
           'pgi',
           'pip',
           'pmix',
           'pnetcdf',
           'python',
           'rdma_core',
           'scif',
           'sensei',
           'slurm_pmi2',
           'ucx',
           'xpmem',
           'yum']

from hpccm.building_blocks.amgx import amgx
from hpccm.building_blocks.apt_get import apt_get
from hpccm.building_blocks.arm_allinea_studio import arm_allinea_studio
from hpccm.building_blocks.boost import boost
from hpccm.building_blocks.catalyst import catalyst
from hpccm.building_blocks.cgns import cgns
from hpccm.building_blocks.charm import charm
from hpccm.building_blocks.cmake import cmake
from hpccm.building_blocks.conda import conda
from hpccm.building_blocks.fftw import fftw
from hpccm.building_blocks.gdrcopy import gdrcopy
from hpccm.building_blocks.generic_autotools import generic_autotools
from hpccm.building_blocks.generic_build import generic_build
from hpccm.building_blocks.generic_cmake import generic_cmake
from hpccm.building_blocks.gnu import gnu
from hpccm.building_blocks.hdf5 import hdf5
from hpccm.building_blocks.hpcx import hpcx
from hpccm.building_blocks.intel_mpi import intel_mpi
from hpccm.building_blocks.intel_psxe import intel_psxe
from hpccm.building_blocks.intel_psxe_runtime import intel_psxe_runtime
from hpccm.building_blocks.julia import julia
from hpccm.building_blocks.knem import knem
from hpccm.building_blocks.kokkos import kokkos
from hpccm.building_blocks.libsim import libsim
from hpccm.building_blocks.llvm import llvm
from hpccm.building_blocks.magma import magma
from hpccm.building_blocks.mkl import mkl
from hpccm.building_blocks.mlnx_ofed import mlnx_ofed
from hpccm.building_blocks.mpich import mpich
from hpccm.building_blocks.multi_ofed import multi_ofed
from hpccm.building_blocks.mvapich2_gdr import mvapich2_gdr
from hpccm.building_blocks.mvapich2 import mvapich2
from hpccm.building_blocks.nccl import nccl
from hpccm.building_blocks.netcdf import netcdf
from hpccm.building_blocks.nsight_compute import nsight_compute
from hpccm.building_blocks.nsight_systems import nsight_systems
from hpccm.building_blocks.nvhpc import nvhpc
from hpccm.building_blocks.nvshmem import nvshmem
from hpccm.building_blocks.ofed import ofed
from hpccm.building_blocks.openblas import openblas
from hpccm.building_blocks.openmpi import openmpi
from hpccm.building_blocks.packages import packages
from hpccm.building_blocks.pgi import pgi
from hpccm.building_blocks.pip import pip
from hpccm.building_blocks.pmix import pmix
from hpccm.building_blocks.pnetcdf import pnetcdf
from hpccm.building_blocks.python import python
from hpccm.building_blocks.rdma_core import rdma_core
from hpccm.building_blocks.scif import scif
from hpccm.building_blocks.sensei import sensei
from hpccm.building_blocks.slurm_pmi2 import slurm_pmi2
from hpccm.building_blocks.ucx import ucx
from hpccm.building_blocks.xpmem import xpmem
from hpccm.building_blocks.yum import yum



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the mpich module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import centos, docker, ubuntu

from hpccm.building_blocks.mpich import mpich
from hpccm.toolchain import toolchain

class Test_mpich(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @ubuntu
    @docker
    def test_defaults_ubuntu(self):
        """Default mpich building block"""
        m = mpich()
        self.assertEqual(str(m),
r'''# MPICH version 3.3.2
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        file \
        gzip \
        make \
        openssh-client \
        perl \
        tar \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://www.mpich.org/static/downloads/3.3.2/mpich-3.3.2.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mpich-3.3.2.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mpich-3.3.2 &&   ./configure --prefix=/usr/local/mpich && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mpich-3.3.2 /var/tmp/mpich-3.3.2.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mpich/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mpich/bin:$PATH''')

    @centos
    @docker
    def test_defaults_centos(self):
        """Default mpich building block"""
        m = mpich()
        self.assertEqual(str(m),
r'''# MPICH version 3.3.2
RUN yum install -y \
        file \
        gzip \
        make \
        openssh-clients \
        perl \
        tar \
        wget && \
    rm -rf /var/cache/yum/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://www.mpich.org/static/downloads/3.3.2/mpich-3.3.2.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mpich-3.3.2.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mpich-3.3.2 &&   ./configure --prefix=/usr/local/mpich && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/mpich-3.3.2 /var/tmp/mpich-3.3.2.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/mpich/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mpich/bin:$PATH''')

    @ubuntu
    @docker
    def test_ldconfig(self):
        """ldconfig option"""
        m = mpich(ldconfig=True, version='3.3')
        self.assertEqual(str(m),
r'''# MPICH version 3.3
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        file \
        gzip \
        make \
        openssh-client \
        perl \
        tar \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://www.mpich.org/static/downloads/3.3/mpich-3.3.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/mpich-3.3.tar.gz -C /var/tmp -z && \
    cd /var/tmp/mpich-3.3 &&   ./configure --prefix=/usr/local/mpich && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    echo "/usr/local/mpich/lib" >> /etc/ld.so.conf.d/hpccm.conf && ldconfig && \
    rm -rf /var/tmp/mpich-3.3 /var/tmp/mpich-3.3.tar.gz
ENV PATH=/usr/local/mpich/bin:$PATH''')

    @ubuntu
    @docker
    def test_runtime(self):
        """Runtime"""
        m = mpich()
        r = m.runtime()
        self.assertEqual(r,
r'''# MPICH
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        openssh-client && \
    rm -rf /var/lib/apt/lists/*
COPY --from=0 /usr/local/mpich /usr/local/mpich
ENV LD_LIBRARY_PATH=/usr/local/mpich/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/mpich/bin:$PATH''')

    def test_toolchain(self):
        """Toolchain"""
        m = mpich()
        tc = m.toolchain
        self.assertEqual(tc.CC, 'mpicc')
        self.assertEqual(tc.CXX, 'mpicxx')
        self.assertEqual(tc.FC, 'mpifort')
        self.assertEqual(tc.F77, 'mpif77')
        self.assertEqual(tc.F90, 'mpif90')



"""This example demonstrates how to specify a multi-stage recipe.

Usage:
$ hpccm.py --recipe recipes/examples/multistage.py
"""

######
# Devel stage
######

# Devel stage base image
Stage0 += baseimage(image='nvcr.io/nvidia/cuda:9.0-devel-ubuntu16.04',
                    _as='devel')

# Install compilers (upstream)
Stage0 += gnu()

# Build FFTW using all default options
Stage0 += fftw()

######
# Runtime stage
######

# Runtime stage base image
Stage1 += baseimage(image='nvcr.io/nvidia/cuda:9.0-runtime-ubuntu16.04')

# Copy the compiler runtime and FFTW from the devel stage and setup the
# runtime environment.
# The _from option is not necessary, but increases clarity.
Stage1 += Stage0.runtime(_from='devel')



r"""
GROMACS 2020

Contents:
  Ubuntu 16.04
  CUDA version 10.1
  GNU compilers (upstream)
  OFED (upstream)
  OpenMPI version 3.1.4
"""

gromacs_version = USERARG.get('gromacs', '2020')

Stage0 += comment(__doc__.strip(), reformat=False)
Stage0 += baseimage(image='nvcr.io/nvidia/cuda:10.1-devel-ubuntu16.04',
                    _as='build')

Stage0 += python(python3=False)

Stage0 += gnu(fortran=False)

Stage0 += cmake(eula=True)

Stage0 += ofed()

Stage0 += openmpi(version='3.1.4')

Stage0 += generic_cmake(cmake_opts=['-D CMAKE_BUILD_TYPE=Release',
                                    '-D CUDA_TOOLKIT_ROOT_DIR=/usr/local/cuda',
                                    '-D GMX_BUILD_OWN_FFTW=ON',
                                    '-D GMX_GPU=ON',
                                    '-D GMX_MPI=ON',
                                    '-D GMX_OPENMP=ON',
                                    '-D GMX_PREFER_STATIC_LIBS=ON',
                                    '-D MPIEXEC_PREFLAGS=--allow-run-as-root'],
                        prefix='/usr/local/gromacs',
                        url='http://ftp.gromacs.org/pub/gromacs/gromacs-{}.tar.gz'.format(gromacs_version))

Stage0 += environment(variables={'PATH': '$PATH:/usr/local/gromacs/bin'})

Stage0 += label(metadata={'gromacs.version': gromacs_version})

######
# Runtime image stage
######
Stage1 += baseimage(image='nvcr.io/nvidia/cuda:10.1-base-ubuntu16.04')

Stage1 += packages(ospackages=['cuda-cufft-10-1'])

Stage1 += Stage0.runtime(_from='build')

Stage1 += environment(variables={'PATH': '$PATH:/usr/local/gromacs/bin'})

Stage1 += label(metadata={'gromacs.version': gromacs_version})



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""PMIX building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import posixpath

import hpccm.config
import hpccm.templates.envvars
import hpccm.templates.ldconfig

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_autotools import generic_autotools
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment

class pmix(bb_base,hpccm.templates.envvars, hpccm.templates.ldconfig):
    """The `pmix` building block configures, builds, and installs the
    [PMIX](https://github.com/openpmix/openpmix) component.

    # Parameters

    annotate: Boolean flag to specify whether to include annotations
    (labels).  The default is False.

    check: Boolean flag to specify whether the `make check` step
    should be performed.  The default is False.

    configure_opts: List of options to pass to `configure`.  The
    default is an empty list.

    disable_FEATURE: Flags to control disabling features when
    configuring.  For instance, `disable_foo=True` maps to
    `--disable-foo`.  Underscores in the parameter name are converted
    to dashes.

    enable_FEATURE[=ARG]: Flags to control enabling features when
    configuring.  For instance, `enable_foo=True` maps to
    `--enable-foo` and `enable_foo='yes'` maps to `--enable-foo=yes`.
    Underscores in the parameter name are converted to dashes.

    environment: Boolean flag to specify whether the environment
    (`CPATH`, `LD_LIBRARY_PATH`, and `PATH`) should be modified to
    include PMIX. The default is True.

    ldconfig: Boolean flag to specify whether the PMIX library
    directory should be added dynamic linker cache.  If False, then
    `LD_LIBRARY_PATH` is modified to include the PMIX library
    directory. The default value is False.

    ospackages: List of OS packages to install prior to configuring
    and building.  For Ubuntu, the default values are `file`, `hwloc`,
    `libevent-dev`, `make`, `tar`, and `wget`. For RHEL-based Linux
    distributions, the default values are `file`, `hwloc`,
    `libevent-devel`, `make`, `tar`, and `wget`.

    prefix: The top level install location.  The default value is
    `/usr/local/pmix`.

    toolchain: The toolchain object.  This should be used if
    non-default compilers or other toolchain options are needed.  The
    default value is empty.

    version: The version of PMIX source to download.  The default value
    is `4.1.2`.

    with_PACKAGE[=ARG]: Flags to control optional packages when
    configuring.  For instance, `with_foo=True` maps to `--with-foo`
    and `with_foo='/usr/local/foo'` maps to
    `--with-foo=/usr/local/foo`.  Underscores in the parameter name
    are converted to dashes.

    without_PACKAGE: Flags to control optional packages when
    configuring.  For instance `without_foo=True` maps to
    `--without-foo`.  Underscores in the parameter name are converted
    to dashes.

    # Examples

    ```python
    pmix(prefix='/opt/pmix/3.1.4', version='3.1.4')
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(pmix, self).__init__(**kwargs)

        self.__baseurl = kwargs.pop('baseurl', 'https://github.com/openpmix/openpmix/releases/download')
        self.__check = kwargs.pop('check', False)
        self.__ospackages = kwargs.pop('ospackages', [])
        self.__prefix = kwargs.pop('prefix', '/usr/local/pmix')
        self.__runtime_ospackages = [] # Filled in by __distro()
        self.__version = kwargs.pop('version', '4.1.2')

        # Set the Linux distribution specific parameters
        self.__distro()

        # Set the environment variables
        self.environment_variables['CPATH'] = '{}:$CPATH'.format(
            posixpath.join(self.__prefix, 'include'))
        self.environment_variables['PATH'] = '{}:$PATH'.format(
            posixpath.join(self.__prefix, 'bin'))
        if not self.ldconfig:
            self.environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(posixpath.join(self.__prefix, 'lib'))

        # Setup build configuration
        self.__bb = generic_autotools(
            annotations={'version': self.__version},
            base_annotation=self.__class__.__name__,
            check=self.__check,
            comment=False,
            devel_environment=self.environment_variables,
            prefix=self.__prefix,
            runtime_environment=self.environment_variables,
            url='{0}/v{1}/pmix-{1}.tar.gz'.format(self.__baseurl,
                                                  self.__version),
            **kwargs)

        # Container instructions
        self += comment('PMIX version {}'.format(self.__version))
        self += packages(ospackages=self.__ospackages)
        self += self.__bb

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['file', 'hwloc', 'libevent-dev', 'make',
                                     'tar', 'wget']
                if self.__check:
                    self.__ospackages.append('perl')
            self.__runtime_ospackages = ['libevent-2.*',
                                         'libevent-pthreads-2.*']
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['file', 'hwloc', 'libevent-devel', 'make',
                                     'tar', 'wget']
                if self.__check:
                    self.__ospackages.append('perl')
            self.__runtime_ospackages = ['libevent']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        p = pmix(...)
        Stage0 += p
        Stage1 += p.runtime()
        ```
        """
        self.rt += comment('PMIX')
        self.rt += packages(ospackages=self.__runtime_ospackages)
        self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""GNU compiler building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion
import posixpath

import hpccm.config
import hpccm.templates.ConfigureMake
import hpccm.templates.envvars
import hpccm.templates.git
import hpccm.templates.ldconfig
import hpccm.templates.rm
import hpccm.templates.tar
import hpccm.templates.wget

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.environment import environment
from hpccm.primitives.shell import shell
from hpccm.toolchain import toolchain

class gnu(bb_base, hpccm.templates.ConfigureMake, hpccm.templates.envvars,
          hpccm.templates.git, hpccm.templates.ldconfig, hpccm.templates.rm,
          hpccm.templates.tar, hpccm.templates.wget):
    """The `gnu` building block installs the GNU compilers from the
    upstream Linux distribution.

    As a side effect, a toolchain is created containing the GNU
    compilers.  The toolchain can be passed to other operations that
    want to build using the GNU compilers.

    # Parameters

    cc: Boolean flag to specify whether to install `gcc`.  The default
    is True.

    configure_opts: List of options to pass to `configure`.  The
    default value is `--disable-multilib`. This option is only
    recognized if a source build is enabled.

    cxx: Boolean flag to specify whether to install `g++`.  The
    default is True.

    environment: Boolean flag to specify whether the environment
    (`LD_LIBRARY_PATH` and `PATH`) should be modified to include
    the GNU compiler. The default is True.

    extra_repository: Boolean flag to specify whether to enable an
    extra package repository containing addition GNU compiler
    packages.  For Ubuntu, setting this flag to True enables the
    `ppa:ubuntu-toolchain-r/test` repository.  For RHEL-based Linux
    distributions, setting this flag to True enables the Software
    Collections (SCL) repository.  The default is False.

    fortran: Boolean flag to specify whether to install `gfortran`.
    The default is True.

    ldconfig: Boolean flag to specify whether the GNU library
    directory should be added dynamic linker cache.  If False, then
    `LD_LIBRARY_PATH` is modified to include the GNU library
    directory. The default value is False. This option is only
    recognized if a source build is enabled.

    openacc: Boolean flag to control whether a OpenACC enabled
    compiler is built. If True, adds `--with-cuda-driver` and
    `--enable-offload-targets=nvptx-none` to the list of host compiler
    `configure` options and also builds the accelerator compiler and
    dependencies (`nvptx-tools` and `nvptx-newlib`). The default value
    is False. This option is only recognized if a source build is
    enabled.

    ospackages: List of OS packages to install prior to configuring
    and building.  For Ubuntu, the default values are `bzip2`, `file`,
    `gcc`, `g++`, `git`, `make`, `perl`, `tar`, `wget`, and
    `xz-utils`.  For RHEL-based Linux distributions, the default
    values are `bzip2`, `file`, `gcc`, `gcc-c++`, `git`, `make`,
    `perl`, `tar`, `wget`, and `xz`. This option is only recognized if
    a source build is enabled.

    prefix: The top level install location.  The default value is
    `/usr/local/gnu`. This option is only recognized if a source build
    is enabled.

    source: Boolean flag to control whether to build the GNU compilers
    from source. The default value is False.

    version: The version of the GNU compilers to install.  Note that
    the version refers to the Linux distribution packaging, not the
    actual compiler version.  For Ubuntu, the version is appended to
    the default package name, e.g., `gcc-7`.  For RHEL-based Linux
    distributions, the version is inserted into the SCL Developer
    Toolset package name, e.g., `devtoolset-7-gcc`.  For RHEL-based
    Linux distributions, specifying the version automatically sets
    `extra_repository` to True.  If a source build is enabled, the
    version is the compiler tarball version on the GNU FTP site and
    the version must be specified. The default is an empty value.

    # Examples

    ```python
    gnu()
    ```

    ```python
    gnu(fortran=False)
    ```

    ```python
    gnu(extra_repository=True, version='7')
    ```

    ```python
    gnu(openacc=True, source=True, version='9.1.0')
    ```

    ```python
    g = gnu()
    openmpi(..., toolchain=g.toolchain, ...)
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(gnu, self).__init__(**kwargs)

        self.__baseurl = kwargs.get('baseurl', 'http://ftpmirror.gnu.org/gcc')
        self.__cc = kwargs.get('cc', True)
        self.configure_opts = kwargs.get('configure_opts',
                                         ['--disable-multilib'])
        self.__cxx = kwargs.get('cxx', True)
        self.__extra_repo = kwargs.get('extra_repository', False)
        self.__fortran = kwargs.get('fortran', True)
        self.__openacc = kwargs.get('openacc', False)
        self.__ospackages = kwargs.get('ospackages', [])
        self.prefix = kwargs.get('prefix', '/usr/local/gnu')
        self.__source = kwargs.get('source', False)
        self.__version = kwargs.get('version', None)
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory

        self.__commands = []       # Filled in below
        self.__compiler_debs = []  # Filled in below
        self.__compiler_rpms = []  # Filled in below
        self.__extra_repo_apt = [] # Filled in below
        self.__runtime_debs = ['libgomp1']
        self.__runtime_rpms = ['libgomp']

        # Output toolchain
        self.toolchain = toolchain()

        if self.__source:
            self.__build()
        else:
            self.__repository()

        # Set the Linux distribution specific parameters
        self.__distro()

        # Fill in container instructions
        self.__instructions()

    def __configure_toolchain_on_build(self):

        directory = posixpath.join(self.prefix, 'bin')

        if self.__cc:
            self.toolchain.CC = posixpath.join(directory, 'gcc')
            self.toolchain.CFLAGS = hpccm.config.get_cpu_optimization_flags('gcc', self.__version)

        if self.__cxx:
            self.toolchain.CXX = posixpath.join(directory, 'g++')
            self.toolchain.CXXFLAGS = hpccm.config.get_cpu_optimization_flags('gcc', self.__version)

        if self.__fortran:
            self.toolchain.FC = posixpath.join(directory, 'gfortran')
            self.toolchain.F77 = posixpath.join(directory, 'gfortran')
            self.toolchain.F90 = posixpath.join(directory, 'gfortran')
            self.toolchain.FFLAGS = hpccm.config.get_cpu_optimization_flags('gcc', self.__version)
            self.toolchain.FCFLAGS = hpccm.config.get_cpu_optimization_flags('gcc', self.__version)

        if "LD_LIBRARY_PATH" in self.environment_variables:
            self.toolchain.LD_LIBRARY_PATH = self.environment_variables["LD_LIBRARY_PATH"]

    def __build(self):
        """Build compilers from source"""

        if not self.__version:
            raise RuntimeError('The compiler version must be specified when performing a source build')

        # Determine which compiler frontends to build
        languages = []
        if self.__cc:
            languages.append('c')
        if self.__cxx:
            languages.append('c++')
        if self.__fortran:
            languages.append('fortran')
        if self.__openacc:
            languages.append('lto')

        # Download source from web
        tarball = 'gcc-{0}.tar.xz'.format(self.__version)
        url = '{0}/gcc-{1}/{2}'.format(self.__baseurl, self.__version, tarball)
        self.__commands.append(self.download_step(url=url, directory=self.__wd))

        # Unpackage
        self.__commands.append(self.untar_step(
            tarball=posixpath.join(self.__wd, tarball),
            directory=self.__wd))

        # Download prerequisites
        self.__commands.append(
            'cd {} && ./contrib/download_prerequisites'.format(
                posixpath.join(self.__wd, 'gcc-{}'.format(self.__version))))

        # Configure accelerator compiler and dependencies
        if self.__openacc:
            # Build nvptx-tools
            # Download
            self.__commands.append(
                self.clone_step(repository='https://github.com/MentorEmbedded/nvptx-tools.git',
                                branch='master', path=self.__wd))

            # Configure
            nvptx_tools = hpccm.templates.ConfigureMake(prefix=self.prefix)
            self.__commands.append(nvptx_tools.configure_step(
                directory=posixpath.join(self.__wd, 'nvptx-tools')))
            # Build
            self.__commands.append(nvptx_tools.build_step())
            self.__commands.append(nvptx_tools.install_step())
            # Cleanup
            self.__commands.append(self.cleanup_step(
                items=[posixpath.join(self.__wd, 'nvptx-tools')]))

            # Setup nvptx-newlib
            self.__commands.append('cd {}'.format(self.__wd))
            self.__commands.append(
                self.clone_step(repository='https://github.com/MentorEmbedded/nvptx-newlib',
                                branch='master', path=self.__wd))
            self.__commands.append('ln -s {0} {1}'.format(
                posixpath.join(self.__wd, 'nvptx-newlib', 'newlib'),
                posixpath.join(self.__wd, 'gcc-{}'.format(self.__version),
                               'newlib')))

            # Accelerator compiler

            # Configure
            accel = hpccm.templates.ConfigureMake(prefix=self.prefix)
            self.__commands.append(accel.configure_step(
                build_directory=posixpath.join(self.__wd, 'accel_objdir'),
                directory=posixpath.join(self.__wd,
                                         'gcc-{}'.format(self.__version)),
                opts=['--enable-languages={}'.format(','.join(languages)),
                      '--target=nvptx-none',
                      '--enable-as-accelerator-for=x86_64-pc-linux-gnu',
                      '--disable-sjlj-exceptions',
                      '--enable-newlib-io-long-long',
                      '--disable-multilib']))

            # Build
            self.__commands.append(accel.build_step())

            # Install
            self.__commands.append(accel.install_step())

        # Configure host compiler
        if self.__openacc:
            self.configure_opts.extend(['--with-cuda-driver=/usr/local/cuda',
                                        '--enable-offload-targets=nvptx-none={}/nvptx-none'.format(self.prefix)])

        self.configure_opts.append('--enable-languages={}'.format(','.join(languages)))

        self.__commands.append(self.configure_step(
            build_directory=posixpath.join(self.__wd, 'objdir'),
            directory=posixpath.join(self.__wd,
                                     'gcc-{}'.format(self.__version))))

        # Build
        self.__commands.append(self.build_step())

        # Install
        self.__commands.append(self.install_step())

        # Environment
        self.environment_variables['PATH'] = '{}:$PATH'.format(
            posixpath.join(self.prefix, 'bin'))
        if self.ldconfig:
            self.__commands.append(self.ldcache_step(
                directory=posixpath.join(self.prefix, 'lib64')))
        else:
            self.environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(posixpath.join(self.prefix, 'lib64'))

        # Setup toolchain
        self.__configure_toolchain_on_build()

        # Cleanup
        self.__commands.append(self.cleanup_step(
            items=[posixpath.join(self.__wd, tarball),
                   posixpath.join(self.__wd, 'gcc-{}'.format(self.__version)),
                   posixpath.join(self.__wd, 'objdir')]))
        if self.__openacc:
            self.__commands.append(self.cleanup_step(
                items=[posixpath.join(self.__wd, 'accel_objdir'),
                       posixpath.join(self.__wd, 'nvptx-newlib')]))

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if self.__source:
            # Build dependencies
            if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
                self.__ospackages = ['bzip2', 'file', 'gcc', 'g++', 'git',
                                     'make', 'perl', 'tar', 'wget', 'xz-utils']
            elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
                self.__ospackages = ['bzip2', 'file', 'gcc', 'gcc-c++', 'git',
                                     'make', 'perl', 'tar', 'wget', 'xz']
            else: # pragma: no cover
                raise RuntimeError('Unknown Linux distribution')

        else:
            # Set libfortran version depending on the Ubuntu version
            if self.__fortran:
                if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
                    if hpccm.config.g_linux_version >= StrictVersion('20.0'):
                        self.__runtime_debs.append('libgfortran5')
                    elif hpccm.config.g_linux_version >= StrictVersion('18.0'):
                        self.__runtime_debs.append('libgfortran4')
                    elif hpccm.config.g_linux_version >= StrictVersion('16.0'):
                        self.__runtime_debs.append('libgfortran3')
                    else: # pragma: no cover
                        raise RuntimeError('Unrecognized Ubuntu version')

            # Setup the environment so that the alternate compiler version
            # is the new default
            if self.__version:
                alternatives = {}
                if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
                    if self.__cc:
                        alternatives['gcc'] = '$(which gcc-{})'.format(
                            self.__version)
                    if self.__cxx:
                        alternatives['g++'] = '$(which g++-{})'.format(
                            self.__version)
                    if self.__fortran:
                        alternatives['gfortran'] = '$(which gfortran-{})'.format(
                            self.__version)
                    alternatives['gcov'] = '$(which gcov-{})'.format(
                        self.__version)
                elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
                    # Default for CentOS 7
                    toolset_path = '/opt/rh/devtoolset-{}/root/usr/bin'.format(
                        self.__version)
                    if hpccm.config.g_linux_version >= StrictVersion('8.0'):
                        # CentOS 8
                        toolset_path = '/opt/rh/gcc-toolset-{}/root/usr/bin'.format(self.__version)

                    if self.__cc:
                        alternatives['gcc'] = posixpath.join(toolset_path, 'gcc')
                    if self.__cxx:
                        alternatives['g++'] = posixpath.join(toolset_path, 'g++')
                    if self.__fortran:
                        alternatives['gfortran'] = posixpath.join(toolset_path,
                                                                  'gfortran')
                    alternatives['gcov'] = posixpath.join(toolset_path, 'gcov')

                else: # pragma: no cover
                    raise RuntimeError('Unknown Linux distribution')

                for tool,alt in sorted(alternatives.items()):
                    self.__commands.append('update-alternatives --install {0} {1} {2} 30'.format(posixpath.join('/usr/bin', tool), tool, alt))

    def __instructions(self):
        """Fill in container instructions"""
        self += comment('GNU compiler')
        if self.__source:
            # Installing from source
            self += packages(ospackages=self.__ospackages)
        else:
            # Installing from package repository
            self += packages(apt=self.__compiler_debs,
                             apt_ppas=self.__extra_repo_apt,
                             release_stream=bool(self.__version), # True/False
                             scl=bool(self.__version), # True / False
                             yum=self.__compiler_rpms)
        if self.__commands:
            self += shell(commands=self.__commands)
        self += environment(variables=self.environment_step())

    def __repository(self):
        """Setup installation from a package repository"""

        if self.__cc:
            self.__compiler_debs.append('gcc')
            self.__compiler_rpms.append('gcc')
            self.toolchain.CC = 'gcc'
            self.toolchain.CFLAGS = hpccm.config.get_cpu_optimization_flags('gcc')

        if self.__cxx:
            self.__compiler_debs.append('g++')
            self.__compiler_rpms.append('gcc-c++')
            self.toolchain.CXX = 'g++'
            self.toolchain.CXXFLAGS = hpccm.config.get_cpu_optimization_flags('gcc')

        if self.__fortran:
            self.__compiler_debs.append('gfortran')
            # libgfortran runtime deb is set is __distro()
            self.__compiler_rpms.append('gcc-gfortran')
            self.__runtime_rpms.append('libgfortran')
            self.toolchain.F77 = 'gfortran'
            self.toolchain.F90 = 'gfortran'
            self.toolchain.FC = 'gfortran'
            self.toolchain.FFLAGS = hpccm.config.get_cpu_optimization_flags('gcc')
            self.toolchain.FCFLAGS = hpccm.config.get_cpu_optimization_flags('gcc')

        # Install an alternate version, i.e., not the default for
        # the Linux distribution
        if self.__version:
            if self.__extra_repo:
                self.__extra_repo_apt = ['ppa:ubuntu-toolchain-r/test']

            # Adjust package names based on specified version
            self.__compiler_debs = [
                '{0}-{1}'.format(x, self.__version)
                for x in self.__compiler_debs]

            if hpccm.config.g_linux_version >= StrictVersion('8.0'):
                # CentOS 8
                self.__compiler_rpms = [
                    'gcc-toolset-{1}-{0}'.format(x, self.__version)
                    for x in self.__compiler_rpms]
            else:
                # CentOS 7
                self.__compiler_rpms = [
                    'devtoolset-{1}-{0}'.format(x, self.__version)
                    for x in self.__compiler_rpms]

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        g = gnu(...)
        Stage0 += g
        Stage1 += g.runtime()
        ```
        """
        self.rt += comment('GNU compiler runtime')
        if self.__source:
            self.rt += copy(_from=_from,
                            dest=posixpath.join(self.prefix, 'lib64'),
                            src=posixpath.join(self.prefix, 'lib64'))
            if self.ldconfig:
                self.rt += shell(commands=[self.ldcache_step(
                    directory=posixpath.join(self.prefix, 'lib64'))])
            else:
                self.rt += environment(variables=self.environment_step(
                    include_only=['LD_LIBRARY_PATH']))
        else:
            self.rt += packages(
                apt=self.__runtime_debs,
                apt_ppas=self.__extra_repo_apt,
                release_stream=bool(self.__version), # True / False
                scl=bool(self.__version), # True / False
                yum=self.__runtime_rpms)

        return str(self.rt)



"""
Recipe for a NVIDIA HPC SDK container and typical development environment.

$ hpccm --recipe nvhpc.py --userarg eula=yes

$ hpccm --recipe nvhpc.py --userarg eula=yes cuda_multi=no arch=x86_64
"""
from distutils.version import StrictVersion
import platform

# Verify correct version of HPCCM is used
if StrictVersion(hpccm.__version__) < StrictVersion('21.7.0'):
  raise RuntimeError('requires HPCCM version 21.7.0 or later')

arch = USERARG.get('arch', platform.machine())
image = USERARG.get('image', 'ubuntu:20.04')
version = USERARG.get('version', '21.7')
release = '20' + version[:2]
default_cuda = '11.4'

cuda_multi=True
if USERARG.get('cuda_multi', False) in [0, '0', 'no', 'false']:
  cuda_multi=False

# The NVIDIA HPC SDK End-User License Agreement must be accepted.
# https://docs.nvidia.com/hpc-sdk/eula
eula=False
if USERARG.get('eula', False) in [1, '1', 'yes', 'true']:
    eula=True
else:
    raise RuntimeError('NVIDIA HPC SDK EULA not accepted. To accept, use "--userarg eula=yes"\nSee NVIDIA HPC SDK EULA at https://docs.nvidia.com/hpc-sdk/eula')

if cuda_multi is True:
    # CUDA 10.2, 11.0, 11.4
    cuda_driver = '"cuda>=10.2"'
else:
    # CUDA 11.4 only, but enhanced forward compatibility means 11.2 is
    # sufficient
    cuda_driver = '"cuda>=11.2"'

Stage0 += baseimage(image=image, _arch=arch)

# Typical development environment
common_packages = ['automake', 'autoconf', 'autoconf-archive', 'binutils',
                   'bzip2', 'ca-certificates', 'cmake', 'diffutils', 'file',
                   'gdb', 'git', 'gzip', 'libtool', 'make', 'patch', 'tar',
                   'vim', 'wget']
Stage0 += packages(apt=common_packages + ['libaec-dev', 'libncursesw5',
                                          'libsz2', 'libtinfo5', 'lmod',
                                          'xz-utils', 'zlib1g-dev'],
                   epel=True,
                   yum=common_packages + ['Lmod', 'libaec-devel', 'xz',
                                          'zlib-devel'])

# Mellanox OFED
Stage0 += mlnx_ofed(version='5.2-2.2.0.0')

# NVIDIA HPC SDK
Stage0 += nvhpc(cuda=default_cuda, cuda_multi=cuda_multi, eula=eula,
                _hpcx=True, mpi=False, version=version)

# Container metadata
Stage0 += environment(variables={'HPCSDK_VERSION': version,
                                 'HPCSDK_RELEASE': release})
Stage0 += label(
    metadata={'com.nvidia.hpcsdk.version': '"{}"'.format(version),
              'com.nvidia.hpcsdk.release': '"{}"'.format(release)})

# nvidia-container-runtime and enroot
Stage0 += environment(variables={
    'MELLANOX_VISIBLE_DEVICES': 'all', # enroot
    'NVIDIA_VISIBLE_DEVICES': 'all',
    'NVIDIA_DRIVER_CAPABILITIES': 'compute,utility',
    'NVIDIA_REQUIRE_CUDA': cuda_driver})



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Label primitive"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import

import hpccm.config

from hpccm.common import container_type

class label(object):
    """The `label` primitive sets container metadata.

    # Parameters

    _app: String containing the [SCI-F](https://www.sylabs.io/guides/2.6/user-guide/reproducible_scif_apps.html)
    identifier.  This also causes the Singularity block to named
    `%applabels` rather than `%labels` (Singularity specific).

    metadata: A dictionary of key / value pairs.  The default is an
    empty dictionary.

    # Examples

    ```python
    label(metadata={'maintainer': 'jane@doe'})
    ```
    """

    def __init__(self, **kwargs):
        """Initialize primitive"""

        #super(label, self).__init__()

        self._app = kwargs.get('_app', '') # Singularity specific
        self.__metadata = kwargs.get('metadata', {})

    def __str__(self):
        """String representation of the primitive"""

        if self.__metadata:
            if hpccm.config.g_ctype == container_type.DOCKER:
                if self._app:
                    logging.warning('The Singularity specific %app.. syntax '
                                    'was requested. Docker does not have an '
                                    'equivalent: using regular LABEL!')

                # Format:
                # LABEL K1=V1 \
                #     K2=V2 \
                #     K3=V3
                keyvals = []
                for key, val in sorted(self.__metadata.items()):
                    keyvals.append('{0}={1}'.format(key, val))

                l = ['LABEL {}'.format(keyvals[0])]
                l.extend(['    {}'.format(x) for x in keyvals[1:]])
                return ' \\\n'.join(l)
            elif hpccm.config.g_ctype == container_type.SINGULARITY:
                # Format:
                # %labels
                #     K1 V1
                #     K2 V2
                #     K3 V3
                keyvals = []
                for key, val in sorted(self.__metadata.items()):
                    keyvals.append('{0} {1}'.format(key, val))

                if self._app:
                    l = ['%applabels {0}'.format(self._app)]
                else:
                    l = ['%labels']
                l.extend(['    {}'.format(x) for x in keyvals])
                return '\n'.join(l)
            elif hpccm.config.g_ctype == container_type.BASH:
                logging.warning('label primitive does not map into bash')
                return ''
            else:
                raise RuntimeError('Unknown container type')
        else:
            return ''

    def merge(self, lst, _app=None):
        """Merge one or more instances of the primitive into a single
        instance.  Due to conflicts or option differences the merged
        primitive may not be exact.

        """

        if not lst: # pragma: nocover
            raise RuntimeError('no items provided to merge')

        labels = {}
        for item in lst:
            if not item.__class__.__name__ == 'label': # pragma: nocover
                logging.warning('item is not the correct type, skipping...')
                continue

            labels.update(item._label__metadata)

        return label(metadata=labels, _app=_app)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""MVAPICH2 building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import posixpath
import re
from copy import copy as _copy

import hpccm.config
import hpccm.templates.envvars
import hpccm.templates.ldconfig
import hpccm.templates.sed

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_autotools import generic_autotools
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment
from hpccm.toolchain import toolchain

class mvapich2(bb_base, hpccm.templates.envvars, hpccm.templates.ldconfig,
               hpccm.templates.sed):
    """The `mvapich2` building block configures, builds, and installs the
    [MVAPICH2](http://mvapich.cse.ohio-state.edu) component.
    Depending on the parameters, the source will be downloaded from
    the web (default) or copied from a source directory in the local
    build context.

    An InfiniBand building block ([OFED](#ofed) or [Mellanox
    OFED](#mlnx_ofed)) should be installed prior to this building
    block.

    As a side effect, a toolchain is created containing the MPI
    compiler wrappers.  The tool can be passed to other operations
    that want to build using the MPI compiler wrappers.

    # Parameters

    annotate: Boolean flag to specify whether to include annotations
    (labels).  The default is False.

    check: Boolean flag to specify whether the `make check` step
    should be performed.  The default is False.

    configure_opts: List of options to pass to `configure`.  The
    default values are `--disable-mcast`.

    cuda: Boolean flag to control whether a CUDA aware build is
    performed.  If True, adds `--enable-cuda --with-cuda` to the list
    of `configure` options, otherwise adds `--disable-cuda`.  If the
    toolchain specifies `CUDA_HOME`, then that path is used, otherwise
    `/usr/local/cuda` is used for the path.  The default value is
    True.

    directory: Path to the unpackaged source directory relative to
    the local build context.  The default value is empty.  If this is
    defined, the source in the local build context will be used rather
    than downloading the source from the web.

    disable_FEATURE: Flags to control disabling features when
    configuring.  For instance, `disable_foo=True` maps to
    `--disable-foo`.  Underscores in the parameter name are converted
    to dashes.

    enable_FEATURE[=ARG]: Flags to control enabling features when
    configuring.  For instance, `enable_foo=True` maps to
    `--enable-foo` and `enable_foo='yes'` maps to `--enable-foo=yes`.
    Underscores in the parameter name are converted to dashes.

    environment: Boolean flag to specify whether the environment
    (`LD_LIBRARY_PATH` and `PATH`) should be modified to include
    MVAPICH2. The default is True.

    gpu_arch: The GPU architecture to use.  Older versions of MVAPICH2
    (2.3b and previous) were hard-coded to use "sm_20".  This option
    has no effect on more recent MVAPICH2 versions.  The default value
    is to use the MVAPICH2 default.

    ldconfig: Boolean flag to specify whether the MVAPICH2 library
    directory should be added dynamic linker cache.  If False, then
    `LD_LIBRARY_PATH` is modified to include the MVAPICH2 library
    directory. The default value is False.

    ospackages: List of OS packages to install prior to configuring
    and building.  For Ubuntu, the default values are `byacc`, `file`,
    `flex`, `make`, `openssh-client`, and `wget`.  For RHEL-based
    Linux distributions, the default values are `byacc`, `file`,
    `flex`, `make`, `openssh-clients`, and `wget`.

    prefix: The top level install location.  The default value is
    `/usr/local/mvapich2`.

    toolchain: The toolchain object.  This should be used if
    non-default compilers or other toolchain options are needed.  The
    default is empty.

    version: The version of MVAPICH2 source to download.  This value
    is ignored if `directory` is set.  The default value is `2.3.4`.

    with_PACKAGE[=ARG]: Flags to control optional packages when
    configuring.  For instance, `with_foo=True` maps to `--with-foo`
    and `with_foo='/usr/local/foo'` maps to
    `--with-foo=/usr/local/foo`.  Underscores in the parameter name
    are converted to dashes.

    without_PACKAGE: Flags to control optional packages when
    configuring.  For instance `without_foo=True` maps to
    `--without-foo`.  Underscores in the parameter name are converted
    to dashes.

    # Examples

    ```python
    mvapich2(cuda=False, prefix='/opt/mvapich2/2.3a', version='2.3a')
    ```

    ```python
    mvapich2(directory='sources/mvapich2-2.3b')
    ```

    ```python
    n = nvhpc(eula=True)
    mvapich2(toolchain=n.toolchain)
    ```

    ```python
    mvapich2(configure_opts=['--disable-fortran', '--disable-mcast'])
    ```
    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(mvapich2, self).__init__(**kwargs)

        self.__baseurl = kwargs.pop('baseurl',
                                    'http://mvapich.cse.ohio-state.edu/download/mvapich/mv2')
        self.__configure_opts = kwargs.pop('configure_opts', ['--disable-mcast'])
        self.__cuda = kwargs.pop('cuda', True)
        self.__gpu_arch = kwargs.pop('gpu_arch', None)
        self.__ospackages = kwargs.pop('ospackages', [])
        self.__preconfigure = []
        self.__prefix = kwargs.pop('prefix', '/usr/local/mvapich2')
        self.__runtime_ospackages = [] # Filled in by __distro()
        # Input toolchain, i.e., what to use when building
        # Create a copy of the toolchain so that it can be modified
        self.__toolchain = _copy(kwargs.pop('toolchain', toolchain()))
        self.__version = kwargs.pop('version', '2.3.4')

        # MVAPICH2 does not accept F90
        self.toolchain_control = {'CC': True, 'CXX': True, 'F77': True,
                                  'F90': False, 'FC': True}

        # Output toolchain
        self.toolchain = toolchain(CC='mpicc', CXX='mpicxx', F77='mpif77',
                                   F90='mpif90', FC='mpifort')

        # Set the configure options
        self.__configure()

        # Set the Linux distribution specific parameters
        self.__distro()

        # Setup the environment variables
        # Set library path
        self.environment_variables['PATH'] = '{}:$PATH'.format(
            posixpath.join(self.__prefix, 'bin'))
        self.runtime_environment_variables['PATH'] = '{}:$PATH'.format(
            posixpath.join(self.__prefix, 'bin'))
        if not self.ldconfig:
            self.environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(posixpath.join(self.__prefix, 'lib'))
            self.runtime_environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(posixpath.join(self.__prefix, 'lib'))
        if self.__cuda:
            # Workaround for using compiler wrappers in the build stage
            self.environment_variables['PROFILE_POSTLIB'] = '"-L{} -lnvidia-ml -lcuda"'.format('/usr/local/cuda/lib64/stubs')

        # Setup build configuration
        self.__bb = generic_autotools(
            annotations={'version': self.__version},
            base_annotation=self.__class__.__name__,
            comment=False,
            configure_opts=self.__configure_opts,
            devel_environment=self.environment_variables,
            preconfigure=self.__preconfigure,
            prefix=self.__prefix,
            runtime_environment=self.runtime_environment_variables,
            toolchain=self.__toolchain,
            url='{0}/mvapich2-{1}.tar.gz'.format(self.__baseurl,
                                                 self.__version),
            **kwargs)

        # Container instructions
        self += comment('MVAPICH2 version {}'.format(self.__version))
        self += packages(ospackages=self.__ospackages)
        self += self.__bb

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['byacc', 'file', 'flex', 'make',
                                     'openssh-client', 'wget']
            self.__runtime_ospackages = ['openssh-client']
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['byacc', 'file', 'flex', 'make',
                                     'openssh-clients', 'wget']
            self.__runtime_ospackages = ['openssh-clients']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def __set_gpu_arch(self):
        """Older versions of MVAPICH2 (2.3b and previous) were hard-coded to
        use the "sm_20" GPU architecture.  Use the specified value
        instead."""

    def __configure(self):
        """Construct the series of shell commands, i.e., fill in
           self.__commands"""

        # MVAPICH2 does not accept F90
        self.__toolchain.F90 = ''

        # Workarounds when using the NV compilers
        if (self.__toolchain.CC and re.match('.*nvc', self.__toolchain.CC) and
            not self.__toolchain.CFLAGS):
            self.__toolchain.CFLAGS = '-fpic -DPIC'
            self.__configure_opts.append('ac_cv_c_compiler_gnu=no')
        if (self.__toolchain.F77 and
            re.match('.*nvfortran', self.__toolchain.F77) and
            not self.__toolchain.FFLAGS):
            self.__toolchain.FFLAGS = '-fpic -DPIC'
        if (self.__toolchain.FC and
            re.match('.*nvfortran', self.__toolchain.FC) and
            not self.__toolchain.FCFLAGS):
            self.__toolchain.FCFLAGS = '-fpic -DPIC'

        # CUDA
        if self.__cuda:
            cuda_home = "/usr/local/cuda"
            if self.__toolchain.CUDA_HOME:
                cuda_home = self.__toolchain.CUDA_HOME

            # The PGI compiler needs some special handling for CUDA.
            # http://mvapich.cse.ohio-state.edu/static/media/mvapich/mvapich2-2.0-userguide.html#x1-120004.5
            if self.__toolchain.CC and re.match('.*pgcc', self.__toolchain.CC):
                self.__configure_opts.append(
                    '--enable-cuda=basic --with-cuda={}'.format(cuda_home))

                # Work around issue when using PGI 19.4
                self.__configure_opts.append('--enable-fast=O1')

                if not self.__toolchain.CFLAGS:
                    self.__toolchain.CFLAGS = '-ta=tesla:nordc'

                if not self.__toolchain.CPPFLAGS:
                    self.__toolchain.CPPFLAGS = '-D__x86_64 -D__align__\(n\)=__attribute__\(\(aligned\(n\)\)\) -D__location__\(a\)=__annotate__\(a\) -DCUDARTAPI='

                if not self.__toolchain.LD_LIBRARY_PATH:
                    self.__toolchain.LD_LIBRARY_PATH = posixpath.join(
                        cuda_home, 'lib64', 'stubs') + ':$LD_LIBRARY_PATH'
            else:
                if not self.__toolchain.LD_LIBRARY_PATH:
                    self.__toolchain.LD_LIBRARY_PATH = posixpath.join(
                        cuda_home, 'lib64', 'stubs') + ':$LD_LIBRARY_PATH'
                self.__configure_opts.append(
                    '--enable-cuda --with-cuda={}'.format(cuda_home))

            # Workaround for using compiler wrappers in the build stage
            self.__preconfigure.append('ln -s {0} {1}'.format(
                posixpath.join(cuda_home, 'lib64', 'stubs', 'libnvidia-ml.so'),
                posixpath.join(cuda_home, 'lib64', 'stubs',
                               'libnvidia-ml.so.1')))
            self.__preconfigure.append('ln -s {0} {1}'.format(
                posixpath.join(cuda_home, 'lib64', 'stubs', 'libcuda.so'),
                posixpath.join(cuda_home, 'lib64', 'stubs', 'libcuda.so.1')))

            # Older versions of MVAPICH2 (2.3b and previous) were
            # hard-coded to use the "sm_20" GPU architecture.  Use the
            # specified value instead.
            if self.__gpu_arch:
                self.__preconfigure.append(
                    self.sed_step(file='Makefile.in',
                                  patterns=[r's/-arch sm_20/-arch {}/g'.format(self.__gpu_arch)]))

        else:
            self.__configure_opts.append('--disable-cuda')

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        m = mvapich2(...)
        Stage0 += m
        Stage1 += m.runtime()
        ```
        """
        self.rt += comment('MVAPICH2')
        # TODO: move the definition of runtime ospackages
        self.rt += packages(ospackages=self.__runtime_ospackages)
        self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Runscript primitive"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import shlex
from six.moves import shlex_quote

import logging # pylint: disable=unused-import

import hpccm.config

from hpccm.common import container_type

class runscript(object):
    """The `runscript` primitive specifies the commands to be invoked
    when the container starts.

    # Parameters

    _args: Boolean flag to specify whether `"$@"` should be appended
    to the command.  If more than one command is specified, nothing is
    appended regardless of the value of this flag.  The default is
    True (Singularity specific).

    _app: String containing the [SCI-F](https://www.sylabs.io/guides/2.6/user-guide/reproducible_scif_apps.html)
    identifier.  This also causes the Singularity block to named `%apprun`
    rather than `%runscript` (Singularity specific).

    commands: A list of commands to execute.  The default is an empty
    list.

    _exec: Boolean flag to specify whether `exec` should be inserted
    to preface the final command.  The default is True (Singularity
    specific).

    # Examples

    ```python
    runscript(commands=['cd /workdir', 'source env.sh'])
    ```

    ```python
    runscript(commands=['/usr/local/bin/entrypoint.sh'])
    ```

    """

    def __init__(self, **kwargs):
        """Initialize primitive"""

        #super(wget, self).__init__()

        self._args = kwargs.get('_args', True) # Singularity specific
        self._app = kwargs.get('_app', '') # Singularity specific
        self._exec = kwargs.get('_exec', True) # Singularity specific
        self.commands = kwargs.get('commands', [])

    def __str__(self):
        """String representation of the primitive"""
        if self.commands:
            if hpccm.config.g_ctype == container_type.DOCKER:
                if self._app:
                    logging.warning('The Singularity specific %app.. syntax was '
                                    'requested. Docker does not have an '
                                    'equivalent: using regular ENTRYPOINT!')

                if len(self.commands) > 1:
                    logging.warning('Multiple commands given to runscript. '
                                    'Docker ENTRYPOINT supports just one cmd: '
                                    'ignoring remaining commands!')
                # Format:
                # ENTRYPOINT ["cmd1", "arg1", "arg2", ...]
                s = []
                s.extend('"{}"'.format(shlex_quote(x))
                    for x in shlex.split(self.commands[0]))
                return 'ENTRYPOINT [' + ', '.join(s) + ']'

            elif hpccm.config.g_ctype == container_type.SINGULARITY:
                if self._exec:
                    # prepend last command with exec
                    self.commands[-1] = 'exec {0}'.format(self.commands[-1])
                if len(self.commands) == 1 and self._args:
                    # append "$@" to singleton command
                    self.commands[0] = '{} "$@"'.format(self.commands[0])
                # Format:
                # %runscript
                #     cmd1
                #     cmd2
                #     exec cmd3
                if self._app:
                    s = ['%apprun {0}'.format(self._app)]
                else:
                    s = ['%runscript']
                s.extend(['    {}'.format(x) for x in self.commands])
                return '\n'.join(s)
            elif hpccm.config.g_ctype == container_type.BASH:
                logging.warning('runscript primitive does not map into bash')
                return ''
            else:
                raise RuntimeError('Unknown container type')
        else:
            return ''

    def merge(self, lst, _app=None):
        """Merge one or more instances of the primitive into a single
        instance.  Due to conflicts or option differences the merged
        primitive may not be exact.

        """

        if not lst: # pragma: nocover
            raise RuntimeError('no items provided to merge')

        cmds = []
        for item in lst:
            if not item.__class__.__name__ == 'runscript': # pragma: nocover
                logging.warning('item is not the correct type, skipping...')
                continue

            cmds.extend(item.commands)

        return runscript(commands=cmds, _app=_app)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""Boost building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import re
import posixpath

import hpccm.config
import hpccm.templates.envvars
import hpccm.templates.ldconfig
import hpccm.templates.rm
import hpccm.templates.tar
import hpccm.templates.wget

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.environment import environment
from hpccm.primitives.shell import shell

class boost(bb_base, hpccm.templates.envvars, hpccm.templates.ldconfig,
            hpccm.templates.rm, hpccm.templates.tar, hpccm.templates.wget):
    """The `boost` building block downloads and installs the
    [Boost](https://www.boost.org) component.

    # Parameters

    b2_opts: List of options to pass to `b2`.  The default is an empty
    list.

    bootstrap_opts: List of options to pass to `bootstrap.sh`.  The
    default is an empty list.

    environment: Boolean flag to specify whether the environment
    (`LD_LIBRARY_PATH`) should be modified to include Boost. The
    default is True.

    ldconfig: Boolean flag to specify whether the Boost library
    directory should be added dynamic linker cache.  If False, then
    `LD_LIBRARY_PATH` is modified to include the Boost library
    directory. The default value is False.

    ospackages: List of OS packages to install prior to building.  For
    Ubuntu, the default values are `bzip2`, `libbz2-dev`, `tar`,
    `wget`, and `zlib1g-dev`.  For RHEL-based Linux distributions the
    default values are `bzip2`, `bzip2-devel`, `tar`, `wget`, `which`,
    and `zlib-devel`.

    prefix: The top level installation location.  The default value
    is `/usr/local/boost`.

    python: Boolean flag to specify whether Boost should be built with
    Python support.  If enabled, the Python C headers need to be
    installed (typically this can be done by adding `python-dev` or
    `python-devel` to the list of OS packages).  This flag is ignored
    if `bootstrap_opts` is set.  The default is False.

    sourceforge: Boolean flag to specify whether Boost should be
    downloaded from SourceForge rather than the current Boost
    repository.  For versions of Boost older than 1.63.0, the
    SourceForge repository should be used.  The default is False.

    version: The version of Boost source to download.  The default
    value is `1.76.0`.

    # Examples

    ```python
    boost(prefix='/opt/boost/1.67.0', version='1.67.0')
    ```

    ```python
    boost(sourceforge=True, version='1.57.0')
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(boost, self).__init__(**kwargs)

        self.__b2_opts = kwargs.get('b2_opts', [])
        self.__baseurl = kwargs.get('baseurl',
                                    'https://boostorg.jfrog.io/artifactory/main/release/__version__/source')
        self.__bootstrap_opts = kwargs.get('bootstrap_opts', [])
        self.__ospackages = kwargs.get('ospackages', [])
        self.__parallel = kwargs.get('parallel', '$(nproc)')
        self.__prefix = kwargs.get('prefix', '/usr/local/boost')
        self.__python = kwargs.get('python', False)
        self.__sourceforge = kwargs.get('sourceforge', False)
        self.__version = kwargs.get('version', '1.76.0')

        self.__commands = [] # Filled in by __setup()
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory

        if self.__sourceforge:
            self.__baseurl = 'https://sourceforge.net/projects/boost/files/boost/__version__'

        # Set the Linux distribution specific parameters
        self.__distro()

        # Construct the series of steps to execute
        self.__setup()

        # Fill in container instructions
        self.__instructions()

    def __instructions(self):
        """Fill in container instructions"""

        self += comment('Boost version {}'.format(self.__version))
        self += packages(ospackages=self.__ospackages)
        self += shell(commands=self.__commands)
        self += environment(variables=self.environment_step())

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['bzip2', 'libbz2-dev', 'tar', 'wget',
                                     'zlib1g-dev']
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['bzip2', 'bzip2-devel', 'tar', 'wget',
                                     'which', 'zlib-devel']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def __setup(self):
        """Construct the series of shell commands, i.e., fill in
           self.__commands"""

        # The download URL has the version format with underscores so
        # pull apart the full version to get the individual
        # components.
        match = re.match(r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<revision>\d+)',
                         self.__version)
        v_underscore = '{0}_{1}_{2}'.format(match.groupdict()['major'],
                                            match.groupdict()['minor'],
                                            match.groupdict()['revision'])

        tarball = 'boost_{}.tar.bz2'.format(v_underscore)
        url = '{0}/{1}'.format(self.__baseurl, tarball)
        url = url.replace('__version__', self.__version)

        # Python support requires pyconfig.h which is not part of the
        # standard Python install.  It requires the development
        # package, python-dev or python-devel.  So skip Python unless
        # it's specifically enabled.
        if not self.__bootstrap_opts and not self.__python:
            self.__bootstrap_opts.append('--without-libraries=python')

        # Download source from web
        self.__commands.append(self.download_step(url=url, directory=self.__wd))
        self.__commands.append(self.untar_step(
            tarball=posixpath.join(self.__wd, tarball), directory=self.__wd))

        # Configure
        self.__commands.append(
            'cd {} && ./bootstrap.sh --prefix={} {}'.format(
                posixpath.join(self.__wd, 'boost_{}'.format(v_underscore)),
                self.__prefix,
                ' '.join(self.__bootstrap_opts)))

        # Build and install
        self.__b2_opts.append('-j{}'.format(self.__parallel))
        self.__b2_opts.append('-q install')
        self.__commands.append('./b2 {0}'.format(' '.join(self.__b2_opts)))

        # Set library path
        libpath = posixpath.join(self.__prefix, 'lib')
        if self.ldconfig:
            self.__commands.append(self.ldcache_step(directory=libpath))
        else:
            self.environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(libpath)

        # Cleanup tarball and directory
        self.__commands.append(self.cleanup_step(
            items=[posixpath.join(self.__wd, tarball),
                   posixpath.join(self.__wd,
                                  'boost_{}'.format(v_underscore))]))

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        b = boost(...)
        Stage0 += b
        Stage1 += b.runtime()
        ```
        """
        self.rt += comment('Boost')
        self.rt += copy(_from=_from, src=self.__prefix, dest=self.__prefix)
        if self.ldconfig:
            self.rt += shell(commands=[self.ldcache_step(
                directory=posixpath.join(self.__prefix, 'lib'))])
        self.rt += environment(variables=self.environment_step())
        return str(self.rt)



"""
Demonstrate how to include a recipe in another recipe
"""

# the gnu-devel recipe file must be in the same directory
hpccm.include('gnu-devel.py')

Stage0 += generic_build(branch='v1.3',
                        build=['make COMPILER=GNU'],
                        install=['cp clover_leaf /usr/local/bin'],
                        repository='https://github.com/UK-MAC/CloverLeaf_Serial.git')
